package edu.lehigh.swat.hawk.core;

import java.util.Iterator;
import java.util.Vector;
import java.io.Serializable;

public class AxiomNode implements Serializable{
	private int joinType;
	private ObjectId atomicNode;
	private Vector<AxiomNode> children;

	private int minCardinality, maxCardinality;
	private AxiomNode allValuesFrom, someValuesFrom, hasValue, onProperty;
	private String restrictionId;

	public AxiomNode(){
		joinType = -10;
		atomicNode = null;
		children = new Vector<AxiomNode>();
		allValuesFrom = null;
		someValuesFrom = null;
		hasValue = null;
		onProperty = null;
		minCardinality = 0;
		maxCardinality = -1;
		restrictionId = null;
	}
	public int getJoinType() {
		return joinType;
	}

	public void setJoinType(int type){
		joinType = type;
		atomicNode = null;
		if(children!=null)
			children = new Vector<AxiomNode>();

		allValuesFrom = null;
		someValuesFrom = null;
		hasValue = null;
		onProperty = null;
		minCardinality = 0;
		maxCardinality = -1;
		restrictionId = null;
	}

	public ObjectId getAtomicNode(){
		return atomicNode;
	}

	public void setAtomicNode(ObjectId aNode){
		atomicNode = aNode;
		joinType = Ontology.ATOMICNODE;
		children = null;

		allValuesFrom = null;
		someValuesFrom = null;
		hasValue = null;
		onProperty = null;
		minCardinality = 0;
		maxCardinality = -1;
		restrictionId = null;
	}

	public Vector<AxiomNode> getChildren(){
		return children;
	}

	public void addChild(AxiomNode child){
		children.add(child);
	}

	public void addChild(Vector<AxiomNode> children){
		this.children.addAll(children);
	}
	public void setRestriction(PropertyRestriction restriction){
		atomicNode = null;
		children = null;
		if(restriction instanceof edu.lehigh.swat.hawk.storage.simplememory.PropertyRestriction){
			minCardinality = ((edu.lehigh.swat.hawk.storage.simplememory.PropertyRestriction)restriction).
								getMinCardinality();
			maxCardinality = ((edu.lehigh.swat.hawk.storage.simplememory.PropertyRestriction)restriction).
								getMaxCardinality();
			restrictionId = ((edu.lehigh.swat.hawk.storage.simplememory.PropertyRestriction)
							restriction).getID().getId();

			if(minCardinality > 0 || maxCardinality > 0)
				joinType = Ontology.CARDINALITYRESTRICTION;


			AxiomNode temp;

			ObjectId oid = ((edu.lehigh.swat.hawk.storage.simplememory.PropertyRestriction)
					restriction).getOnProperty();
			if(oid!=null){
				temp = new AxiomNode();
				temp.setAtomicNode(oid);
				this.setOnProperty(temp);
			}
			oid = ((edu.lehigh.swat.hawk.storage.simplememory.PropertyRestriction)
					restriction).getSomeValuesFrom();
			if(oid!=null){
				temp = new AxiomNode();
				temp.setAtomicNode(oid);
				this.setSomeValuesFrom(temp);
				joinType = Ontology.SOMEVALUERESTRICTION;
			}
			oid = ((edu.lehigh.swat.hawk.storage.simplememory.PropertyRestriction)
					restriction).getAllValuesFrom();
			if(oid!=null){
				temp = new AxiomNode();
				temp.setAtomicNode(oid);
				this.setAllValuesFrom(temp);
				joinType = Ontology.ALLVALUERESTRICTION;
			}

			oid = ((edu.lehigh.swat.hawk.storage.simplememory.PropertyRestriction)
					restriction).getHasValue();
			if(oid!=null){
				temp = new AxiomNode();
				temp.setAtomicNode(oid);
				temp.setHasValue(temp);
				joinType = Ontology.HASVALUERESTRICTION;
			}
		}
	}

	/**
	 * Get the property restricted on.
	 */
	public AxiomNode getOnProperty(){
		return onProperty;
	}

	/**
	 * Set the property restricted on.
	 */
	public void setOnProperty(AxiomNode p){
		onProperty = p;
	}

	/**
	 * Get the specified restriction class of 'allValuesFrom'.
	 */
	public AxiomNode getAllValuesFrom(){
		return allValuesFrom;
	}

	/**
	 * Set the specified restriction class of 'allValuesFrom'.
	 */
	public void setAllValuesFrom(AxiomNode c){
		allValuesFrom = c;
	}

	/**
	 * Get the specified restriction class of 'someValuesFrom'.
	 */
	public AxiomNode getSomeValuesFrom(){
		return someValuesFrom;
	}

	/**
	 * Set the specified restriction class of 'someValuesFrom'.
	 */
	public void setSomeValuesFrom(AxiomNode c){
		someValuesFrom = c;
	}

	public String getRestrictionId(){
		return restrictionId;
	}
	/**
	 * Get the specified number of 'minCardinality'.
	 */
	public int getMinCardinality(){
		return minCardinality;
	}

	/**
	 * Set the specified number of 'minCardinality'.
	 */
	public void setMinCardinality(int n){
		minCardinality = n;
	}

	/**
	 * Get the specified number of 'maxCardinality'.
	 */
	public int getMaxCardinality(){
		return maxCardinality;
	}

	/**
	 * Set the specified number of 'maxCardinality'.
	 */
	public void setMaxCardinality(int n){
		maxCardinality = n;
	}

	/**
	 * Get the specified value of 'hasValue'.
	 */
	public AxiomNode getHasValue(){
		return hasValue;
	}

	/**
	 * Set the specified value of 'hasValue'.
	 */
	public void setHasValue(AxiomNode c){
		hasValue = c;
	}

	public boolean equals(Object o){
		AxiomNode node;
		if(o instanceof AxiomNode )
			node = (AxiomNode)o;
		else
			return false;

		if(this.getJoinType() == Ontology.ATOMICNODE && node.getJoinType() == Ontology.ATOMICNODE){
			if(this.getAtomicNode().getId().equals(node.getAtomicNode().getId()))
				return true;
			else
				return false;
		}else if(this.getJoinType() > Ontology.ATOMICNODE && node.getJoinType() > Ontology.ATOMICNODE){
			if(this.getJoinType() != node.getJoinType())
				return false;
			else{
				for(Iterator<AxiomNode> it = this.getChildren().iterator();it.hasNext();){
					AxiomNode c = it.next();
					if(!node.getChildren().contains(c))
						return false;
				}
				return true;
			}
		}else if(this.getRestrictionId()!=null && node.getRestrictionId()!=null){
			if(this.getRestrictionId().equals(node.getRestrictionId()))
				return true;
			else
				return false;
		}

		return false;
	}
}
