/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/


package edu.lehigh.swat.hawk.core;


//import java.util.Iterator;

/**
 * A general representation of the Classes.
 * implementations depend on the storage models
 * @author Z. Pan
 * @version 1.0
 */

public interface OntClass extends OntObject{


  /**
   * add a direct subClass
   */
  public void addSubClass(ObjectId c);

  /**
   * get the direct subClasses
   */
  public ObjectId[] getDirectSubClasses();

  /**
   * delete a direct subClass
   */
  public void deleteSubClass(ObjectId c);

  /**
   * add a direct super Class
   */
  public void addSuperClass(ObjectId c);

  /**
   * get the direct super Classes
   */
  public ObjectId[] getDirectSuperClasses();

  /**
   * delete a direct super Class
   */
  public void deleteSuperClass(ObjectId c);

  /**
   * Add a property restriction.
   */
  public void addPropertyRestriction(ObjectId r);

  /**
   * Delete a property restriction.
   */
  public void deletePropertyRestriction(ObjectId r);

  /**
   * Get the property restrictions.
   */
  public ObjectId[] getPropertyRestrictions();

  /**
   * Add a equivalent class.
   */
  public void addEquivalentClass(ObjectId ec);

  /**
   * Delete a equivalent class.
   */
  public void deleteEquivalentClass(ObjectId r);

  /**
   * Get the equivalent classes.
   */
  public ObjectId[] getEquivalentClasses();

  /**
   * Add a 'disjointWith' class.
   */
  public void addDisjointWith(ObjectId id);

  /**
   * Delete a 'disjointWith' class.
   */
  public void deleteDisjointWith(ObjectId id);

  /**
   * Get the disjointWith classes.
   */
  public ObjectId[] getDisjointWith();

  /**
   * Add an 'UnionOf' class.
   */
  public void addUnionOf(ObjectId u);

  /**
   * Delete an 'UnionOf' class.
   */
  public void deleteUnionOf(ObjectId u);

  /**
   * Get the 'IntersectionOf' classes.
   */
  public ObjectId[] getUnionOfs();

  public AxiomNode[] getUnionAxiomNodes();

  /**
   * Add an 'IntersectionOf' class or restriction.
   */
  public void addIntersectionOf(ObjectId i);

  /**
   * Delete an 'IntersectionOf' class or restriction.
   */
  public void deleteIntersectionOf(ObjectId i);

  /**
   * Get the 'IntersectionOf' classes or restrictions.
   */
  public ObjectId[] getIntersectionOfs();

  public AxiomNode[] getIntersectionAxiomNodes();

  /**
   * Add an 'OneOf' object.
   */
  public void addOneOf(ObjectId one);

  /**
   * Delete an 'OneOf' object.
   */
  public void deleteOneOf(ObjectId one);

  /**
   * Get the 'OneOf' objects.
   */
  public ObjectId[] getOneOfs();

  /**
   * Add an 'ComplementOf' class or restriction.
   */
  public void addComplementOf(ObjectId i);

  /**
   * Delete an 'ComplementOf' class or restriction.
   */
  public void deleteComplementOf(ObjectId i);

  /**
   * Get the 'ComplementOf' classes or restrictions.
   */
  public ObjectId[] getComplementOfs();

  /**
   * Test if this class is deprecated.
   */
  public boolean isDeprecated();

  /**
   * Set the class to be deprecated.
   */
  public void setDeprecated(boolean dep);

  /**
   * Set the satisfiability of this class.
   */
  public void setSatisfiable(Boolean sat);

  /**
   * Test if this class is satisfiable.
   */
  public Boolean isSatisfiabele();

}
