package edu.lehigh.swat.hawk.helpers;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.Reader;
import java.io.StringReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintStream;

import java.sql.Date;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;

import edu.lehigh.swat.hawk.core.DatatypeValue;
import edu.lehigh.swat.hawk.core.DocId;
import edu.lehigh.swat.hawk.core.ObjectId;
import edu.lehigh.swat.hawk.storage.ResultSet;
import edu.lehigh.swat.hawk.storage.simpledb.DatabaseStorage;
import edu.lehigh.swat.hawk.storage.simpledb.SparqlQuery;


//zhp2 on Nov.12, 2007: If you use filter on some variable, you better select that varibale as distinguished variable!
public class SparqlQueryDB {

	  public static DatabaseStorage storage;
	  public static final int OUT = 0x01, ERR = 0x02;
      public String ontoName = "";
	  public String ontoReference = "";
	  public HashMap ontoReferenceMap = new HashMap();
	  public String perspective = "";
	  public int perspectiveSeq = -1;
      protected static PrintStream streamErr = null, streamOut = null;
      protected static boolean _verbose = false;

      public SparqlQueryDB() {
     	 streamErr = System.err;
     	 streamOut = System.out;
      }

      public SparqlQueryDB(PrintStream stream, int streamType) {
     	 if (SparqlQueryDB.ERR == (streamType & SparqlQueryDB.ERR)) {
     		 streamErr = stream;
     	 }
     	 if (SparqlQueryDB.OUT == (streamType & SparqlQueryDB.OUT)) {
     		 streamOut = stream;
     	 }
      }

      public SparqlQueryDB(PrintStream err, PrintStream out) {
     	 streamErr = err;
     	 streamOut = out;
      }

      public static void setStreamErr(PrintStream err) {
     	 SparqlQueryDB.streamErr = err;
      }

      public static void setStreamOut(PrintStream out) {
     	 SparqlQueryDB.streamOut = out;
      }

      public static PrintStream getStreamErr() {
     	 return SparqlQueryDB.streamErr;
      }

      public static PrintStream getStreamOut() {
     	 return SparqlQueryDB.streamOut;
      }

	  public static void main(String[] args) {
	    if(args.length!=2){
	      SparqlQueryDB.streamErr.println("usage: java SparqlQueryDB DB_File_Name query_File_Name");
	      System.exit(0);
	    }

	    SparqlQueryDB test = new SparqlQueryDB();

	    storage = new DatabaseStorage(args[0]);
	  //  storage.setReasonerUrl("http://128.188.123.123:8080");
	    storage.open();
	    ResultSet rs = null;
	    SparqlQuery qry = null;

	    ArrayList queryList = test.parseQuery(args[1]);
	    //SparqlQueryDB.streamOut.println("queryList = " + queryList);
	    for(int i = 0; queryList != null && i < queryList.size();i++){
	    	qry = (SparqlQuery)queryList.get(i);
	    	SparqlQueryDB.streamOut.println("\n-----------------\nQuery " + (i+1) + ":");
	    	//SparqlQueryDB.streamOut.println(qry.getSql());
	        long startTime = System.currentTimeMillis();
	        rs = storage.issueQuery(qry);
	        long time = System.currentTimeMillis()-startTime;
	        SparqlQueryDB.streamOut.println("Time used: "+time+" ms.");

	        //String s = rs.toString();
	        int n = rs.getResultsNumber();
	        SparqlQueryDB.streamOut.println("\n------------------\nTotal number of results: "+n);

//	        while(rs.hasMoreResults()){
//                    System.out.println(rs.getNextResult().toString());
//            }
	    }

	    storage.close();
	  }
	  public void setStorage(DatabaseStorage strg){
		  storage = strg;
	  }

	  /**
	   * Parse the query from file.
	   */
	  public ArrayList parseQuery(String fn){
	    SQueryConfigParser qcp = new SQueryConfigParser(fn);
	    qcp.parse();
	    return qcp.queryList;
	  }

      public ArrayList parseQueryString (String query) {
     	 StringReader reader = new StringReader(query);
     	 SQueryConfigParser qcp = new SQueryConfigParser(reader);
          qcp.parse();
          return qcp.queryList;
      }

	  /////////////////////////////////////////////////////
	  ///code below copied from ubt.test.java ////////////
	  ////////////////////////////////////////////////////
	  abstract private class ConfigParser {
	    public LineNumberReader reader_;
	    protected static final char C_SEC_BEGIN = '[';
	    protected static final char C_SEC_END = '}';
	    private String fileName_;

	    public ConfigParser(String fileName) {
	      fileName_ = fileName;
	    }

	    public void parse() {
	      String line;

	      handleBeginOfFile();
	      while ((line = readLine()) != null) {
	        handleLine(line);
	      }
	      handleEndOfFile();
	    }

	    abstract protected void handleLine(String line);
	    protected void handleBeginOfFile() {/*do nothing*/}
	    protected void handleEndOfFile() {/*do nothing*/}

	    private String readLine() {
	      String line = null;

	      try {
	        while ((line = reader_.readLine()) != null) {
	          line.trim();
	          if (line.length() > 0) break;
	        }
	      }
	      catch (IOException e) {
	        e.printStackTrace();
	        System.exit(1);
	      }

	      return line;
	    }
        protected void outPrintln(String str) {
     	   if (SparqlQueryDB.streamOut != null) {
     		   SparqlQueryDB.streamOut.println(str);
     	   }
     	   else {
     		   outPrintln(str);
     	   }
        }

        protected void errPrintln(String str) {
     	   if (SparqlQueryDB.streamErr != null) {
     		   SparqlQueryDB.streamErr.println(str);
     	   }
     	   else {
     		   errPrintln(str);
     	   }
        }

	    protected void error(String error) {
	      outPrintln(fileName_ + ": Line " + reader_.getLineNumber() + ": " + error);
	      System.exit(1);
	    }

	    protected boolean isStartOfSection(String line) {
	      if (line.startsWith("SELECT")) {
	        return true;
	      }
	      return false;
	    }
	    protected boolean isEndOfSection(String line) {
		      if (line.endsWith("}")) {
		        return true;
		      }
		      return false;
		    }
	  	}


	  private class SQueryConfigParser extends ConfigParser {

	    protected static final char C_ATOM_BEGIN = '(';
	    protected static final char C_ATOM_END = ')';
	    protected static final char C_ARG_DELIM = ' ';
	    protected static final char C_WHERE_BEGIN = '{';
	    protected static final char C_WHERE_END = '}';

	    public SparqlQuery query_; //current section
	    public ArrayList queryList;
	    public HashMap varToUri = new HashMap();
        StringBuffer selectClause = new StringBuffer();
        StringBuffer whereClause = new StringBuffer();
        StringBuffer fromClause = new StringBuffer();
        HashMap variablesHash = new HashMap();
        HashMap selectVars  = new HashMap();
        HashMap hash = new HashMap();
        boolean insideWhere = false;
        boolean inFilter = false;
        boolean isOrderByClause = false;
        ArrayList triplesList = new ArrayList();
        int triple = 0;
        boolean endOfSection = false;
        HashMap tables = new HashMap();
        HashMap properties = new HashMap();
        HashMap classes = new HashMap();
        ArrayList vcol = new ArrayList();
        String[] selectVariables = null;
        String[] columnNames = null;
        int count = 0;

	    public SQueryConfigParser(String CONFIG_FILE_QUERY) {
	      super(CONFIG_FILE_QUERY);
	      try {
	        reader_ = new LineNumberReader(new FileReader(CONFIG_FILE_QUERY));
	        query_ = null;
	        queryList = new ArrayList();
	      }
	      catch (FileNotFoundException e) {
	        e.printStackTrace();
	        System.exit(1);
	      }
	    }

	   	public SQueryConfigParser(Reader rdr) {
	   		 super(null);
	   		 reader_ = new LineNumberReader(rdr);
	   		 query_ = null;
	   		 queryList = new ArrayList();
	   	 }

	    protected void handleLine(String line) {
	      if (line.startsWith("#")) return;

	      int n;

	      Hashtable propertyTypes = storage.getPropertyTypes();

	      if (line.startsWith("[")){
	    	 perspective = line.substring(1, line.length()-1);
	    	 //outPrintln("perspective: " + perspective);
	    	 perspectiveSeq = storage.getOntSeq(perspective);
	    	 outPrintln("perspectiveSeq: " + perspectiveSeq);
	      }else if(line.startsWith("PREFIX")){
	    	//  if(line.contains("owl")){
	    		  String[] prefixToks = line.split(" ");
	    		  ontoName = prefixToks[1];
	    		  ontoReference = prefixToks[2];
	    		  //remove < >
	    		  ontoReference = ontoReference.substring(1, ontoReference.length()-1);
	    		  ontoReferenceMap.put(ontoName, ontoReference);
	    		//  outPrintln("ontoName = " + ontoName);
	    	//  }
	      }else {
	    	  //Only meant to deal with SELECT lines
	    	  if (isStartOfSection(line)) {
		        endPrevSection();
		        endOfSection = false;
		        //new section
		        query_ = new SparqlQuery();
		        varToUri = new HashMap();
		        String did = line.substring(1, line.length() - 1);
		       // query_.id_ = line.substring(1, line.length() - 1);
		        if (did.length() <= 0) {
		          error("Empty Id!");
		        }
		        query_.setPerspective(new DocId(did));
		        query_.addToSql("SELECT DISTINCT ");
		      }
	    	  if(isEndOfSection(line)){
		    	  endOfSection = true;
		      }
		      String[] tokens = line.split(" ");
		      String[] filterTokens = null;
		      int countSelects = 0;

		      for(n = 0; n < tokens.length; n++){
		    	  if(tokens[n] != null && !"".equals(tokens[n]) && tokens[n].charAt(0) == '?' && !insideWhere){
		    		  String tok = tokens[n].substring(0);
		    		  String[] vars = tok.split(",");	// Ameet 7 Nov 2007
		    		  /*if(tok.endsWith(",")){
		    			  tok = tok.substring(0, tok.length()-1);
		    		  }*/
		    		  for(int i=0;i<vars.length;i++)
		    			  selectVars.put(vars[i], new Integer(countSelects++));
		    		  //selectVars.put(tok, new Integer(countSelects++));
		    		  outPrintln("putting into selectVars: " + selectVars);
		    	  } else if("WHERE".equals(tokens[n])){
		    		  insideWhere = true;
		    		  if(!tokens[n+1].startsWith("{")){
		    			  error("Syntax error!, Expected { after WHERE");
		    		  }
		    	  } else if("FILTER".equals(tokens[n])){
		    		  inFilter = true;
		    	  } else if("ORDER".equals(tokens[n])){
		    		  isOrderByClause = true;
		    	  }
		    	  if(insideWhere && !inFilter && !isOrderByClause) {
		    		  //first trim off excess chars at end of token
		    		  String token = tokens[n];
		    		  if(token.endsWith(".}")){
		    			  token = token.substring(0, token.length()-2);
		    		  }else if(token.endsWith(".") || token.endsWith("}")){
		    			  token = token.substring(0, token.length()-1);
		    		  } else if(token.startsWith("{")){
		    			  token = token.substring(1);
		    		  }
		    		  if(!token.equals("WHERE") && !token.equals("WHERE{") && !"".equals(token) && !" ".equals(token)){
		    		  //then put the triples into hashmaps and put each triple into the list
			    		  outPrintln("triple: " + triple);
			    		  outPrintln("token: '" + token + "'");
			    		  if(token.startsWith("\"") && !(token.endsWith("\"") || token.endsWith("\"}"))){
			    			  while(!token.endsWith("\"") && !token.endsWith("\"}") && !token.endsWith("\".")){
			    				  System.out.println("tokens[n+1] = " + tokens[n+1]);
			    				  token = token + " " + tokens[n+1];
			    				  n++;
			    			  }
				    		//  token = token.substring(1, token.length()-1);
				    		  if(token.endsWith("}") || token.endsWith("."))
				    			  token = token.substring(0, token.length()-1);
				    		  outPrintln("token: '" + token + "'");
			    		  }
			    		  if(triple == 0){
			    			  hash.put("subject", token);
			    			  triple++;
			    		  } else if(triple == 1){
			    			  hash.put("predicate", token);
			    			  triple++;
			    		  } else if(triple == 2){
			    			  hash.put("object", token);
			    			  outPrintln("hash: " + hash);
			    			  triplesList.add(hash);
			    			  triple = 0;
			    			  hash = new HashMap();
			    		  }
		    	  	  }
		    		  //START OF FILTER SECTION OF QUERY
		    		  //parsing of the filter occurs after the triples list has been dealt with
		    	  }	else if (inFilter && !isOrderByClause){
		    		  int numTokensLeft = tokens.length - n;
		    		  filterTokens = new String[numTokensLeft];
		    		  for (int k = 0; k < numTokensLeft; k++){
		    			  filterTokens[k] = tokens[n++];
		    		  }
		    	  } else if(isOrderByClause){

		    	  }
		      }
		      //THIS IS WHERE THE MAJOR TRANSLATION OCCURS --i.e. if you are seeing an error in translation, look here first!
		      String subject = "";
		      String predicate = "";
		      String object = "";
		      HashMap hash = null;
		      int ptype = 0;
		      String tn = "";
		      System.out.println("selectVars before switch statement: " + selectVars);
		      System.out.println("selectVars.size(): " + selectVars.size());
		      if(selectVariables == null)
		    	  selectVariables = new String[selectVars.size()];
		      if(columnNames == null)
		    	  columnNames = new String[selectVars.size()];

		      outPrintln("triplesList = " + triplesList.size());
              if(triplesList == null)
                  System.exit(0);
              int tlsize = triplesList.size();
		      for(int i = 0;  i < tlsize; i++){
                  hash = (HashMap)triplesList.get(i);
		    	  subject = hash.get("subject").toString();
		    	  predicate = hash.get("predicate").toString();
		    	  object = hash.get("object").toString();
		    //	  if(subject.startsWith("<")){
		    //		  subject = subject.substring(1, subject.length()-1);
		    //	  } else if(object.startsWith("<")){
		    //		  object = object.substring(1, object.length()-1);
		    //	  }

		    	  ptype = 0;
		    	  if(predicate.contains("type")){
		    		  ptype = 1;
		    	  }
		    	  outPrintln("ptype: " + ptype);
		    	  outPrintln("hash: " + hash);

		    	  switch(ptype){
		    	  case 1:
		    		  String arg = object;
		    		  String arg2 = subject;

		    		  if(arg.contains("?") && !subject.startsWith("?")){
		    			  arg = subject;
		    			  arg2 = object;
		    		  }
		    		  ObjectId objid;
		              if(arg.indexOf('#')>0||arg.indexOf('/')>0)
		                objid = new ObjectId(arg);
		              else {
		            	int index = arg.indexOf(":");
		            	objid = new ObjectId(ontoReferenceMap.get(arg.substring(0, index+1)) + arg.substring(index+1));
		              }
		              outPrintln("objid: " + objid.getId());

		              tn = storage.getTableName(objid);
		              if(perspectiveSeq != -1){
		            	  tn = tn+"_"+perspectiveSeq;
		              }
                      if(!tn.equals("thing_owl"))
		                tn = tn + "_view";

                      String vn = tn;
                      if(tables.containsKey(tn))  //zhp2
                           vn = tn+count;      //zhp2 on 11/12/07
            //          if(!storage.getDBName().contains("mysql"))
                           fromClause.append(vn).append(" AS " + tn).append(", ");
             //         else
             //              fromClause.append(vn).append(", ");

		              outPrintln("adding to fromClause:"  + fromClause);
		              tables.put(vn,"class");
		              varToUri.put(arg2, "ui" + count + ".uri");
		              if(arg.equals(object))
		            	  variablesHash.put(arg2, vn + ".object");
		              else
		            	  variablesHash.put(arg2, vn + ".subject");
		              if(properties.get(arg)==null){

		                if(selectVars.containsKey(arg2)){
                             fromClause.append("uri_index AS ui" + count + ", ");
		                	//cew - 7/24/07 - making select variables be added in correct order
		                	int placement = new Integer(selectVars.get(arg2).toString()).intValue();
		                	columnNames[placement] = arg2.substring(1, arg2.length());	// Ameet 7 Nov 2007
		                	//columnNames[placement] = arg;
		                	selectVariables[placement] = "ui" + count+ ".uri, ";
		                //	selectClause.append("count" + count+ ".uri, ");
		                	selectVars.remove(arg2);
                            whereClause.append(tn + ".id = ui" + count+ ".id AND ");
		                }
		                vcol.add(arg);
		                classes.put(arg2, vn + ".id");
		                outPrintln("classes after the add: " + classes);
		              }else{
		                whereClause.append(properties.get(arg)).append("=");
		                whereClause.append(vn).append(".id AND ");
		              }
		      break;
		      case 0:
		              ObjectId oid;
		              if(predicate.indexOf('#')>0||predicate.indexOf('/')>0)
		                oid = new ObjectId(predicate);
		              else if(predicate.indexOf(":")>0){
		            	  int index = predicate.indexOf(":");
		            	  oid = new ObjectId(ontoReferenceMap.get(predicate.substring(0, index+1)) + predicate.substring(index+1));
		              } else
		                oid = new ObjectId(perspective+predicate);
		              tn = storage.getTableName(oid)+"_p";
		              outPrintln("oid: " + oid.getId());
		              outPrintln("tn: " + tn);

		              vn = tn;

		              if(perspectiveSeq != -1)
			              	vn = vn+"_"+perspectiveSeq;
                      if(propertyTypes.containsKey(tn)){
                          if(((String)propertyTypes.get(tn)).startsWith("hasinverse"))
                             vn = vn+"_inv";
		               }
                      if(!vn.equals("label_rdfs_p"))
		                  vn = vn+"_view";

		              String varName = vn;
		              if(!tables.containsKey(vn)){
        //                if(!storage.getDBName().contains("mysql")){
                            fromClause.append(vn).append(", ");
                            tables.put(vn, "property");
        /*                }else{
                            varName = vn+count;
                            fromClause.append(varName).append(", ");
                            tables.put(varName,"property");
                          }
		  */            }else {
		            	varName = tn + count;
          //              if(!storage.getDBName().contains("mysql"))
                            fromClause.append(vn).append(" AS " + varName).append(", ");
		    //        	else
            //               fromClause.append(varName).append(", ");
                        tables.put(varName,"property");
		              }

		              //second argument, object.
		              if(object.startsWith("?")){
		            	variablesHash.put(object, varName + ".object");
		                if(classes.get(object)!=null){
		                  whereClause.append(classes.get(object)).append("=");
		                  whereClause.append(varName).append(".object AND ");
		                }
		                if(properties.get(object)!=null){
		                  whereClause.append(properties.get(object)).append("=");
		                  whereClause.append(varName).append(".object AND ");
		                }
		                if(properties.get(object)==null&&classes.get(object) == null){
		                    String select_column_name = varName+".object";
		                    outPrintln("select_column_name: " + select_column_name);
		                    outPrintln("object: " + object);

		                    String type = (String)propertyTypes.get(tn);
		                   // varToUri.put(object, select_column_name);
		                    if(type!=null&&(type.equalsIgnoreCase("datatype")
		                    		|| type.equalsIgnoreCase("date")
		                    		|| type.equalsIgnoreCase("int") ||type.equalsIgnoreCase("float"))){ // Ameet, Sep 5 2007
		                      if(selectVars.containsKey(object)){
                                  varToUri.put(object, select_column_name);
			                      //            	cew - 7/24/07 - making select variables be added in correct order
		                    	  int placement = new Integer(selectVars.get(object).toString()).intValue();
		                    	  columnNames[placement] = object.substring(1, object.length());	// Ameet, 7 Nov 2007
		                    	  //columnNames[placement] = object;
		                    	  selectVariables[placement] = select_column_name + ", ";
		                    	 // selectClause.append(select_column_name).append(", ");
		                    	  selectVars.remove(object);
		                      }
		                      tables.remove(varName);
		                      tables.put(varName, "datatype");
		                    }
		                    else{
		                      if(selectVars.containsKey(object)){

    		                      fromClause.append("uri_index AS uio"+ count +", ");
				                outPrintln("selectVars: " + selectVars);
				                outPrintln("object: " + object);
				                varToUri.put(object, "uio" + count + ".uri");

				                //cew - 7/24/07 - making select variables be added in correct order
				                  int placement = new Integer(selectVars.get(object).toString()).intValue();
				                  columnNames[placement] = object.substring(1, object.length());	// Ameet, 7 Nov 2007
				                  //columnNames[placement] = object;
				                  selectVariables[placement] = "uio" + count+ ".uri, ";
		                    	//  selectClause.append("uio"+ count +".uri, ");
		                    	  selectVars.remove(object);
		                          whereClause.append(select_column_name+" = uio"+ count +".id AND ");
                              }

		                    }
		                    properties.put(object, select_column_name);
		                    vcol.add(object);
		                }
		              }
		              else{// argument is not variable
		            	outPrintln("object is not a variable: " + object);
		                //cew - 7/25/07 -fix for variable type, to deal with literal, numerics and uris correctly
		                if(object.startsWith("<") && object.endsWith(">")){
		                	fromClause.append("uri_index AS uio"+ count +", ");
		                	object = object.substring(1, object.length()-1);
		                	whereClause.append("uio"+ count +".uri='").append(object).append("' AND ");
		                	 whereClause.append(varName).append(".object=uio"+ count +".id AND ");
		                } else {//if(subject.startsWith("\"") && subject.endsWith("\"")){
		                	String select_column_name = varName + ".object";
		                	whereClause.append(select_column_name).append("=").append(object).append(" AND ");
		                }

                                String type = (String)propertyTypes.get(tn);
                             // varToUri.put(object, select_column_name);
                              if(type!=null&&(type.equalsIgnoreCase("datatype")
                                      || type.equalsIgnoreCase("date")
                                      || type.equalsIgnoreCase("int") ||type.equalsIgnoreCase("float"))){
                                  tables.remove(varName);
                                  tables.put(varName, "datatype");
                            }
		              }

		              //first argument, subject.
		              if(subject.startsWith("?")){
		            	  variablesHash.put(subject, varName + ".subject");
		            	  outPrintln("properties: " + properties);
		            	  outPrintln("classes: " + classes);
		                if (properties.get(subject) != null) {
		                  whereClause.append(properties.get(subject)).append("=");
		                  whereClause.append(varName).append(".subject ").append(" AND ");
		                }
		                if (classes.get(subject) != null){
		                  whereClause.append(classes.get(subject)).append("=");
		                  whereClause.append(varName).append(".subject AND ");
		                }
		                if(properties.get(subject)==null&&classes.get(subject) == null){
		                  String select_column_name = varName + ".subject";

		                  if(selectVars.containsKey(subject)){
		                    fromClause.append("uri_index AS uis" + count + ", ");
			                outPrintln("selectVars: " + selectVars);
			                outPrintln("subject: " + subject);
			                outPrintln("select_column_name: " + select_column_name);
			                varToUri.put(subject, "uis" + count + ".uri");

			                  //		                	cew - 7/24/07 - making select variables be added in correct order
			                  int placement = new Integer(selectVars.get(subject).toString()).intValue();
			                  columnNames[placement] = subject.substring(1, subject.length());	// Ameet, 7 Nov 2007
			                  //columnNames[placement] = subject;
			                  selectVariables[placement] = "uis" + count + ".uri, ";
		                	  //selectClause.append("uis" + count + ".uri, ");
		                	  selectVars.remove(subject);

                             whereClause.append(select_column_name).append(" = ");
		                     whereClause.append("uis" + count + ".id AND ");
		                  }

		                  vcol.add(subject);
		                  properties.put(subject, select_column_name);
		                  outPrintln("properties:" + properties);
		                }
		              }
		              else{  //argument is not variable
		                  //cew - 7/25/07 -fix for variable type, to deal with literal, numerics and uris correctly
                          if(!subject.startsWith("<")){
                              if(subject.indexOf(":")>0){
                                 int index = subject.indexOf(":");
                                    subject = ontoReferenceMap.get(subject.substring(0, index+1)) + subject.substring(index+1);
                                 }
                              }else if(subject.endsWith(">")){
                                subject = subject.substring(1, subject.length()-1);
                              }
                              fromClause.append("uri_index AS uis"+ count +", ");
                              whereClause.append("uis"+ count +".uri='").append(subject).append("' AND ");
                              whereClause.append(varName).append(".subject=uis"+ count +".id AND ");
		              }
		          }//end switch
		    	  count++;
		      }
		      triplesList = new ArrayList();
		      //END OF MAJOR TRANSLATION AREA

		     //START OF FILTER TRANSLATION
		      if(filterTokens != null){
		    	  for(int t = 0; t < filterTokens.length; t++){
		    		  boolean caseSensitive = true;
		    		  boolean varFound = false;
		    		  String variable = "";
		    		  String varName = "";
		    		  String matches = "";
		    		  String uri = "";

		    		//  outPrintln("inFilter");
		    		//  outPrintln("tokens[n] = " + tokens[n]);
		    		//  outPrintln("varToUri: " + varToUri);

		    		  if("&&".equals(filterTokens[t])){
		    			  uri = varToUri.get(variable).toString();
		    			  varName = variablesHash.get(variable).toString();
						  if(!caseSensitive)
		    				  whereClause.append(uri + " LIKE \"%" + matches.toLowerCase() + "%\" OR " + uri + " LIKE \"%" + matches.toUpperCase() + "%\" AND ");
		    			  else
		    				  whereClause.append(uri + " LIKE \"%" + matches + "%\" AND ");
		    		//	  outPrintln("whereClause = " + whereClause);
		    		  }else if(filterTokens[t].contains("regex")){
		    			  while(t < filterTokens.length && inFilter){
		    	    		//  outPrintln("filterTokens[t] = " + filterTokens[t]);
		    				//  outPrintln("n = " + n);
		    				  if(filterTokens[t].contains("?")){
		    					  int indexof = filterTokens[t].indexOf("?");
		    					  int indexofComma = filterTokens[t].indexOf(",");
		    					  if(indexofComma != -1)
		    						  variable = filterTokens[t].substring(indexof, indexofComma);
		    					  else
		    						  variable = filterTokens[t].substring(indexof);

		    				  } else if(filterTokens[t].contains("\"")){
		    					  String phrase = filterTokens[t];
					    		  if(phrase.startsWith("\"") && !(phrase.endsWith("\"") || phrase.endsWith("\")") ||  phrase.endsWith("\")}"))){
					    			  while(!(phrase.endsWith("\"") || phrase.endsWith("\")") ||  phrase.endsWith("\")}"))){
					    				  phrase = phrase + " " + filterTokens[t+1];
					    				  t++;
					    			  }
						    		  filterTokens[t] = phrase;
					    		  }
		    					  int index1 = filterTokens[t].indexOf("\"");
		    					  int index2 = filterTokens[t].lastIndexOf("\"");
		    					  outPrintln("filterTokens: " + filterTokens[t]);
		    					  String compare = filterTokens[t].substring(index1+1, index2);
		    					  if("i".equals(compare))
		    						  caseSensitive = false;
		    					  else
		    						  matches = compare;
		    				  }
		    				  if(filterTokens[t].contains(")")){
		    					  inFilter = false;
		    					  break;
		    				  }
		    				  t++;
		    			  }
		    			  uri = varToUri.get(variable).toString();

		    			  if(!caseSensitive)
		    				  whereClause.append(uri + " LIKE \"%" + matches.toLowerCase() + "%\" OR " + uri + " LIKE \"%" + matches.toUpperCase() + "%\" AND ");
		    			  else
		    				  whereClause.append(uri + " LIKE \"%" + matches + "%\" AND ");
		    			 // outPrintln("whereClause = " + whereClause);

		    		  } else if (filterTokens[t].contains("(")){
		    			  String comparisonOp = "";
		    			  String value = "";
		    			  while(t < filterTokens.length && inFilter){
		    				  outPrintln("filterTokens = " + filterTokens[t]);
		    				  if("&&".equals(filterTokens[t])){
		    					  varName = variablesHash.get(variable).toString();
		    					  whereClause.append(varName + " " + comparisonOp + " " + value + " AND ");
		    					 // outPrintln("whereClause = " + whereClause);
		    				  }else if(filterTokens[t].contains("?")){
		    					  int indexof = filterTokens[t].indexOf("?");
		    					  int indexofComma = filterTokens[t].indexOf(",");
		    					  if(indexofComma != -1)
		    						  variable = filterTokens[t].substring(indexof, indexofComma);
		    					  else
		    						  variable = filterTokens[t].substring(indexof);
			    			  } else if(filterTokens[t].contains("<") || filterTokens[t].contains(">") || filterTokens[t].contains("=")){
			    				  comparisonOp = filterTokens[t]; //assuming spaces separating for now
			    			  } else {
			    				 // outPrintln("filterTokens[t] = " + filterTokens[t]);
			    				  if(filterTokens[t].endsWith(")"))
			    					  value = filterTokens[t].substring(0, filterTokens[t].length()-1);
			    				  else if(filterTokens[t].endsWith(")}"))
			    					  value = filterTokens[t].substring(0, filterTokens[t].length()-2);
			    				  else
			    					  value = filterTokens[t];
			    				//  outPrintln("filterTokens[t] in else: '" + value + "'");

			    				  boolean isDate = true; // Ameet, Sep 5, 2007
			    				  try{
		            				  Date date = Date.valueOf(value.trim().substring(1, value.trim().length()-1));
		            			  }catch(IllegalArgumentException e){
		            				  isDate = false;
		            			  }

			    				  if(!isDate || !(comparisonOp.equals("<") ||
			    						  comparisonOp.equals(">") ||
			    						  comparisonOp.equals("=") ||
			    						  comparisonOp.equals(">=") ||
			    						  comparisonOp.equals("<=") ||
			    						  comparisonOp.equals("!="))){
			    					  Double dbl = new Double(value.trim());
			    					  //if its a number, save it, otherwise, not sure what it would be
			    					  if(dbl.isNaN()){
			    						  value = "";
			    					  }
			    				  }
			    			  }
		    				  if(filterTokens[t].contains(")")){
		    					  inFilter = false;
		    					  break;
		    				  }
		    				  t++;
		    			  }
		    			  outPrintln("variable: " + variable);
		    			  varName = variablesHash.get(variable).toString();
						  whereClause.append(varName + " " + comparisonOp + " " + value + " AND ");
						//  outPrintln("whereClause = " + whereClause);
		    		  }
		    	  }
		      }
    		  //END OF FILTER TRANSLATION

		      //if we found the "}" in the current line, then query is over, so save the SQL into the query_ object
		      if(endOfSection){
		    	  //cew - 7/24/07 - making select variables be added in correct order
		    	  System.out.println("selectVariables.length = " + selectVariables.length);
		    	  for(int k = 0; k < selectVariables.length; k++){
                      if(selectVariables[k]!=null)  //zhp2 on 7/3/08
		    		  selectClause.append(selectVariables[k]);
		    	  }
		    	  System.out.println("selectClause: " + selectClause.toString());
		    	  query_.setHeaders(columnNames);
		    	  //have to get rid of trailing comma, or trailing AND
			      if(selectClause.length() > 0)
			    	  selectClause = new StringBuffer(selectClause.toString().substring(0, selectClause.toString().length()-2));
			      if(fromClause.length() > 0)
			    	  fromClause = new StringBuffer(" FROM " + fromClause.toString().substring(0, fromClause.toString().length()-2));
			      if(whereClause.length() > 0)
			    	  whereClause = new StringBuffer(" WHERE " + whereClause.toString().substring(0, whereClause.toString().length()-4));
			      query_.addToSql(selectClause.toString());
			      query_.addToSql(fromClause.toString());
			      query_.addToSql(whereClause.toString());
                  query_.setDBTables(tables);            //zhp2 on 11/12/2007
			      selectClause = new StringBuffer();
			      fromClause = new StringBuffer();
			      whereClause = new StringBuffer();
			      selectVariables = null;
			      columnNames = null;
			      count = 0;

		      }
		    }
	    }

	    protected void handleEndOfFile() {
	      if (query_ != null) {
	        endPrevSection();
	      }
	    }

	    private void endPrevSection() {
	      if (query_ != null) {
	    	  System.out.println("adding query: " + query_.getSql());
	    	  queryList.add(query_);
	      }
	      triplesList = new ArrayList();
	      triple=0;
	      insideWhere = false;
	      inFilter = false;
	      isOrderByClause = false;
	      selectVars = new HashMap();
	      tables = new HashMap();
	      properties = new HashMap();
	      classes = new HashMap();
	      vcol = new ArrayList();
	    }
	}
}
