/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/

package edu.lehigh.swat.hawk.owl;


import edu.lehigh.swat.hawk.core.*;

import org.apache.xmlbeans.XmlObject;

import java.io.*;
import java.util.*;
import java.math.BigInteger;

import org.kr.dl.dig.v1_1.*;
import org.kr.dl.dig.reasoner.v1_1.Reasoner;
import org.kr.dl.dig.reasoner.v1_1.impl.HTTPReasoner;

/** Experimental Renderer for the new DIG language. This is primarily
 * intended as a testbed for the new language definition that the <a
 * href="http://dl.kr.org/dig">DIG</a> group are working on. Very
 * similar to the original SHIQ renderer in its makeup. Warning: This
 * may not work properly.

/**
 * OWL to DIG Writer. Output owl ontology to DIG 1.1 format.
 * @author Z. Pan
 * @version 1.0
 */

public class OwlDigWriter {

      /** Create a new instance of the renderer */
      public OwlDigWriter() {
      }

      /** Set options for the renderer.  */
      public void setOption(String str, int i) {
      }

      /* A pair of tables, one mapping names to concepts, the other
         mapping concepts to names. The tables represent the mappings
         that the renderer has used. These are required because names
         get mangled when sending stuff to FaCT. */

      /** Render an ontology in SHIQ using the given writer */
      public TellsDocument writeOntology(Ontology o, Writer w, String kburi) throws Exception {

          TellsDocument tellDoc = TellsDocument.Factory.newInstance();
          tellDoc.addNewTells().setUri(kburi);
          TellsDocument.Tells tells = tellDoc.getTells();

         OntClass[] classes = o.getAllPlusImportedClasses();
         int clen = classes.length;
         for(int i=0;i<clen;i++)
         {
              edu.lehigh.swat.hawk.core.OntClass cl = classes[i];
              String className = cl.getID().getId();
            //  if(!cl.getID().getNameSpace().equals(""))
                tells.addNewDefconcept().setName(className);
          }

          OntProperty[] properties = o.getAllPlusImportedProperties();
          int plen = properties.length;
          for(int i=0;i<plen;i++)
          {
              OntProperty sl = properties[i];
              String propertyName = sl.getID().getId();

              if (sl.isDatatype()==null||!sl.isDatatype().booleanValue()) {
                tells.addNewDefrole().setName(propertyName);
              }
              else {
                tells.addNewDefattribute().setName(propertyName);
              }
          }

         Iterator it = o.getIndividualIterator();
         while(it.hasNext()){
           Individual ind = (Individual)it.next();
           writeIndividual(o, ind, tells);
         }

         writeAxioms(o, tells);

          try {
              tellDoc.save(w);
          } catch (IOException e) {
              throw new Exception(e);
          }
          return tellDoc;
      }

      private void writeIndividual(Ontology ont, Individual ind, TellsDocument.Tells tells){

          String iname = ind.getID().getId();
          //tells.addNewDefindividual().setName(iname);

          ObjectId[] sc = ind.getSourceClasses();
          int sclen = sc.length;
          for(int i=0;i<sclen;i++){
            ObjectId scid = sc[i];
            InstanceofDocument.Instanceof ins = tells.addNewInstanceof();
            ins.addNewIndividual().setName(iname);
            ins.addNewCatom().setName(scid.getId());
          }

          ObjectId[] pps = ind.getProperties();
          int pplen = pps.length;
          for(int i=0;i<pplen;i++){
            ObjectId pp = pps[i];
            ObjectId[] pvs = ind.getValuesForProperty(pp);
            int pvlen = pvs.length;
            for(int j=0;j<pvlen;j++){
              ObjectId pv = pvs[j];
              if(pv instanceof DatatypeValue){
                ValueDocument.Value value = tells.addNewValue();
                value.addNewIndividual().setName(iname);
                value.addNewAttribute().setName(pp.getId());

                DatatypeValue dv = (DatatypeValue)pv;
                if(dv.getType().equals(DatatypeValue.INTEGER))
                  value.setIval(new BigInteger(dv.getValue()));
                else
                  value.setSval(dv.getValue());
              }
              else{
                RelatedDocument.Related related = tells.addNewRelated();
                related.addNewIndividual().setName(iname);
                related.addNewRatom().setName(pp.getId());
                related.addNewIndividual().setName(pv.getId());
              }
            }
        }
      }

      private void writeAxioms(Ontology ont, TellsDocument.Tells tells){
          List axioms = ont.getAllAxioms();
          Iterator<edu.lehigh.swat.hawk.core.Axiom> it = axioms.iterator();
          while(it.hasNext()){
              edu.lehigh.swat.hawk.core.Axiom axiom = it.next();
              AxiomNode left = axiom.getLeftAxiomNode();
              AxiomNode right = axiom.getRightAxiomNode();
              int type = axiom.getType();
              switch(type){
              case Ontology.SUBCLASSOF:
                  ConceptPair pair = tells.addNewImpliesc();
                  writeAxiomNode(left, pair);
                  writeAxiomNode(right, pair);
                  break;
             case Ontology.EQUIVALENTCLASS:
                 pair = tells.addNewEqualc();
                 writeAxiomNode(left, pair);
                 writeAxiomNode(right, pair);
                 break;
             case Ontology.SUBPROPERTYOF:
                RolePair rpair = tells.addNewImpliesr();
                writeAxiomNode(left, rpair);
                writeAxiomNode(right, rpair);
                break;
             case Ontology.EQUIVALENTPROPERTY:
                rpair = tells.addNewImpliesr();
                writeAxiomNode(left, rpair);
                writeAxiomNode(right, rpair);
                break;
             case Ontology.DISJOINTWITH:
                pair = tells.addNewImpliesc();
                writeAxiomNode(left, pair);
                Concept con = pair.addNewNot();
                writeAxiomNode(right, con);
                break;
             default:
                break;

              }
          }
      }

      private void writeAxiomNode(AxiomNode node, Concept parent){
         int type = node.getJoinType();
         if(type==edu.lehigh.swat.hawk.core.Ontology.ATOMICNODE){
                    parent.addNewCatom().setName(node.getAtomicNode().getId());
         }else if(node.getOnProperty()!=null&&type<edu.lehigh.swat.hawk.core.Ontology.ATOMICNODE){
                 if(node.getAllValuesFrom()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.ALLVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewAll();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     writeAxiomNode(node.getAllValuesFrom(),pair);
                 }else if(node.getSomeValuesFrom()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.SOMEVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewSome();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     writeAxiomNode(node.getSomeValuesFrom(),pair);
                 }else if(node.getHasValue()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.HASVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewSome();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.addNewIset().addNewIndividual().setName(node.getHasValue().getAtomicNode().getId());
                 }else if(node.getMaxCardinality()!=PropertyRestriction.INFINITY){
                     NumRoleConceptPair pair = parent.addNewAtmost();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.setNum(BigInteger.valueOf(node.getMaxCardinality()));
                     pair.addNewTop();
                 }
                 if(node.getMinCardinality()>0){
                     NumRoleConceptPair pair = parent.addNewAtleast();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.setNum(BigInteger.valueOf(node.getMinCardinality()));
                     pair.addNewTop();
                 }
          }else{
             switch (type) {
             case Ontology.INTERSECTIONOF:
                 Concepts cons = parent.addNewAnd();
                 List children = node.getChildren();
                 Iterator<AxiomNode> it = children.iterator();
                 while (it.hasNext()) {
                     AxiomNode child = it.next();
                     writeAxiomNode(child, cons);
                 }
                 break;

             case Ontology.UNIONOF:
                cons = parent.addNewOr();
                children = node.getChildren();
                it = children.iterator();
                while (it.hasNext()) {
                    AxiomNode child = it.next();
                    writeAxiomNode(child, cons);
                }
                 break;

             case Ontology.COMPLEMENTOF:
                 Concept con = parent.addNewNot();
                 children = node.getChildren();
                 it = children.iterator();
                 if (it.hasNext()) {
                     AxiomNode child = it.next();
                     writeAxiomNode(child, con);
                 }
                 break;

             default:
                 break;
             }
         }
      }

      private void writeAxiomNode(AxiomNode node, RoleConceptPair parent){
         int type = node.getJoinType();
         if(type==edu.lehigh.swat.hawk.core.Ontology.ATOMICNODE){
                    parent.addNewCatom().setName(node.getAtomicNode().getId());
         }else if(node.getOnProperty()!=null&&type<edu.lehigh.swat.hawk.core.Ontology.ATOMICNODE){
                 if(node.getAllValuesFrom()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.ALLVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewAll();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     writeAxiomNode(node.getAllValuesFrom(),pair);
                 }else if(node.getSomeValuesFrom()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.SOMEVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewSome();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     writeAxiomNode(node.getSomeValuesFrom(),pair);
                 }else if(node.getHasValue()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.HASVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewSome();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.addNewIset().addNewIndividual().setName(node.getHasValue().getAtomicNode().getId());
                 }else if(node.getMaxCardinality()!=PropertyRestriction.INFINITY){
                     NumRoleConceptPair pair = parent.addNewAtmost();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.setNum(BigInteger.valueOf(node.getMaxCardinality()));
                     pair.addNewTop();
                 }
                 if(node.getMinCardinality()>0){
                     NumRoleConceptPair pair = parent.addNewAtleast();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.setNum(BigInteger.valueOf(node.getMinCardinality()));
                     pair.addNewTop();
                 }
          }else{
             switch (type) {
             case Ontology.INTERSECTIONOF:
                 Concepts cons = parent.addNewAnd();
                 List children = node.getChildren();
                 Iterator<AxiomNode> it = children.iterator();
                 while (it.hasNext()) {
                     AxiomNode child = it.next();
                     writeAxiomNode(child, cons);
                 }
                 break;

             case Ontology.UNIONOF:
                cons = parent.addNewOr();
                children = node.getChildren();
                it = children.iterator();
                while (it.hasNext()) {
                    AxiomNode child = it.next();
                    writeAxiomNode(child, cons);
                }
                 break;

             case Ontology.COMPLEMENTOF:
                 Concept con = parent.addNewNot();
                 children = node.getChildren();
                 it = children.iterator();
                 if (it.hasNext()) {
                     AxiomNode child = it.next();
                     writeAxiomNode(child, con);
                 }
                 break;

             default:
                 break;
             }
         }
      }

      private void writeAxiomNode(AxiomNode node, Concepts parent){
         int type = node.getJoinType();
         if(type==edu.lehigh.swat.hawk.core.Ontology.ATOMICNODE){
                    parent.addNewCatom().setName(node.getAtomicNode().getId());
         }else if(node.getOnProperty()!=null&&type<edu.lehigh.swat.hawk.core.Ontology.ATOMICNODE){
                 if(node.getAllValuesFrom()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.ALLVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewAll();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     writeAxiomNode(node.getAllValuesFrom(),pair);
                 }else if(node.getSomeValuesFrom()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.SOMEVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewSome();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     writeAxiomNode(node.getSomeValuesFrom(),pair);
                 }else if(node.getHasValue()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.HASVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewSome();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.addNewIset().addNewIndividual().setName(node.getHasValue().getAtomicNode().getId());
                 }else if(node.getMaxCardinality()!=PropertyRestriction.INFINITY){
                     NumRoleConceptPair pair = parent.addNewAtmost();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.setNum(BigInteger.valueOf(node.getMaxCardinality()));
                     pair.addNewTop();
                 }
                 if(node.getMinCardinality()>0){
                     NumRoleConceptPair pair = parent.addNewAtleast();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.setNum(BigInteger.valueOf(node.getMinCardinality()));
                     pair.addNewTop();
                 }
          }else{
             switch (type) {
             case Ontology.INTERSECTIONOF:
                 Concepts cons = parent.addNewAnd();
                 List children = node.getChildren();
                 Iterator<AxiomNode> it = children.iterator();
                 while (it.hasNext()) {
                     AxiomNode child = it.next();
                     writeAxiomNode(child, cons);
                 }
                 break;

             case Ontology.UNIONOF:
                cons = parent.addNewOr();
                children = node.getChildren();
                it = children.iterator();
                while (it.hasNext()) {
                    AxiomNode child = it.next();
                    writeAxiomNode(child, cons);
                }
                 break;

             case Ontology.COMPLEMENTOF:
                 Concept con = parent.addNewNot();
                 children = node.getChildren();
                 it = children.iterator();
                 if (it.hasNext()) {
                     AxiomNode child = it.next();
                     writeAxiomNode(child, con);
                 }
                 break;

             default:
                 break;
             }
         }
      }

      private void writeAxiomNode(AxiomNode node, ConceptPair parent){
         int type = node.getJoinType();
         if(type==edu.lehigh.swat.hawk.core.Ontology.ATOMICNODE){
                    parent.addNewCatom().setName(node.getAtomicNode().getId());
         }else if(node.getOnProperty()!=null&&type<edu.lehigh.swat.hawk.core.Ontology.ATOMICNODE){
                 if(node.getAllValuesFrom()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.ALLVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewAll();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     writeAxiomNode(node.getAllValuesFrom(),pair);
                 }else if(node.getSomeValuesFrom()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.SOMEVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewSome();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     writeAxiomNode(node.getSomeValuesFrom(),pair);
                 }else if(node.getHasValue()!=null&&type==edu.lehigh.swat.hawk.core.Ontology.HASVALUERESTRICTION){
                     RoleConceptPair pair = parent.addNewSome();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.addNewIset().addNewIndividual().setName(node.getHasValue().getAtomicNode().getId());
                 }else if(node.getMaxCardinality()!=PropertyRestriction.INFINITY){
                     NumRoleConceptPair pair = parent.addNewAtmost();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.setNum(BigInteger.valueOf(node.getMaxCardinality()));
                     pair.addNewTop();
                 }
                 if(node.getMinCardinality()>0){
                     NumRoleConceptPair pair = parent.addNewAtleast();
                     pair.addNewRatom().setName(node.getOnProperty().getAtomicNode().getId());
                     pair.setNum(BigInteger.valueOf(node.getMinCardinality()));
                     pair.addNewTop();
                 }
          }else{
             switch (type) {
             case Ontology.INTERSECTIONOF:
                 Concepts cons = parent.addNewAnd();
                 List children = node.getChildren();
                 Iterator<AxiomNode> it = children.iterator();
                 while (it.hasNext()) {
                     AxiomNode child = it.next();
                     writeAxiomNode(child, cons);
                 }
                 break;

             case Ontology.UNIONOF:
                cons = parent.addNewOr();
                children = node.getChildren();
                it = children.iterator();
                while (it.hasNext()) {
                    AxiomNode child = it.next();
                    writeAxiomNode(child, cons);
                }
                 break;

             case Ontology.COMPLEMENTOF:
                 Concept con = parent.addNewNot();
                 children = node.getChildren();
                 it = children.iterator();
                 if (it.hasNext()) {
                     AxiomNode child = it.next();
                     writeAxiomNode(child, con);
                 }
                 break;

             default:
                 break;
             }
         }
      }

      private void writeAxiomNode(AxiomNode node, RolePair parent){
          int type = node.getJoinType();
          if(type==edu.lehigh.swat.hawk.core.Ontology.ATOMICNODE){
                  parent.addNewRatom().setName(node.getAtomicNode().getId());
          }else if(type==Ontology.INVERSEOF){
              List children = node.getChildren();
              Iterator<AxiomNode> it = children.iterator();
              if (it.hasNext()) {
                  AxiomNode child = it.next();
                  String pname = child.getAtomicNode().getId();
                  parent.addNewInverse().addNewRatom().setName(pname);
              }
          }
      }

}




