/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/


package edu.lehigh.swat.hawk.owl;
//import edu.lehigh.swat.hawk.storage.OntWriter;
import java.io.OutputStream;
import java.util.*;

import com.hp.hpl.jena.vocabulary.OWL;
import edu.lehigh.swat.hawk.core.*;
import edu.lehigh.swat.hawk.storage.simplememory.Ontology;

import org.apache.xerces.dom.DocumentImpl;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.Element;
import org.w3c.dom.Text;

/**
 * OWL writer, output OWL document in RDF format.
 * @author Z. Pan
 * @version 1.0
 */

public class OwlRdfWriter extends OntWriter{

  /**
   * Constructor
   */
  public OwlRdfWriter() {
  }

  /**
   * write a document to output
   */
  public void write(OutputStream out, edu.lehigh.swat.hawk.core.Document doc) throws Exception{
    output = out;
    xmlBase = doc.getXMLBase();
    ns = doc.getNameSpaces();

    document = new DocumentImpl();
    Element root = document.createElement("rdf:RDF");
    if(xmlBase!=null&&!xmlBase.equals(""))
      root.setAttribute("xml:base",xmlBase);
    document.appendChild(root);

    Element ontology = document.createElement("owl:Ontology");
    ontology.setAttribute("rdf:about","");
    //imports
    DocId[] imps = doc.getImports();
    if(imps!=null){
      int len = imps.length;
      for(int i=0;i<len;i++){
        Element eimp = document.createElement("owl:imports");
        eimp.setAttribute("rdf:resource",imps[i].getId());
        ontology.appendChild(eimp);
      }
    }

    Iterator it = doc.getAllDescriptions();
    while(it.hasNext()){
      String comment = (String)it.next();
      Element ecomment = document.createElement("rdfs:comment");
      Text txt = document.createTextNode(comment);
      ecomment.appendChild(txt);
      ontology.appendChild(ecomment);
    }

    String label = doc.getLabel();
    if(label!=null){
      Element elabel = document.createElement("rdfs:label");
      Text txt = document.createTextNode(label);
      elabel.appendChild(txt);
      ontology.appendChild(elabel);
    }

    root.appendChild(ontology);

    if(doc instanceof Ontology){
      Ontology ont = (Ontology)doc;

      DocId[] priors = ont.getPriorOnts();
      if(priors!=null){
        int len = priors.length;
        Element eimp = document.createElement("owl:priorVersion");
        for(int i=0;i<len;i++){
          Element eprior = document.createElement("owl:Ontology");
          eprior.setAttribute("rdf:about",priors[i].getId());
          eimp.appendChild(eprior);
        }
        ontology.appendChild(eimp);
      }

      DocId[] compatibles = ont.getCompatiblePriorOnts();
      if(compatibles!=null){
        int len = compatibles.length;
        Element eimp = document.createElement("owl:backwardCompatibleWith");
        for(int i=0;i<len;i++){
          eimp.setAttribute("rdf:resource",compatibles[i].getId());
        }
        ontology.appendChild(eimp);
      }

      Vector<Axiom> axioms = (Vector<Axiom>)ont.getAllAxioms();
      ArrayList<AxiomNode> temp = new ArrayList<AxiomNode>();
      for(Iterator<Axiom> it1 = axioms.iterator();it1.hasNext();){
    	  Axiom axiom = it1.next();
    	  if(!temp.contains(axiom.getLeftAxiomNode())){
    		  temp.add(axiom.getLeftAxiomNode());
    		  Element c = writeEntity(ont, axiom.getLeftAxiomNode());
    		  root.appendChild(c);
    	  }
      }
    } //end if ontology

    Individual[] individuals = doc.getAllIndividuals();
    int ilen = individuals.length;
    for(int i=0;i<ilen;i++){
      Element ind = writeIndividual(doc, individuals[i]);
      root.appendChild(ind);
    }

    //different
    Collection[] distinct = doc.getDistinctIndividualsSets();
    int dlen = distinct.length;
    for(int i=0;i<dlen;i++){
      if(distinct[i].size()<=2)
        continue;
      Element diff = document.createElement("owl:AllDifferent");
      Element dism = document.createElement("owl:distinctMembers");
      dism.setAttribute("rdf:parseType","Collection");
      Iterator ci = distinct[i].iterator();
      while(ci.hasNext()){
        ObjectId oid = (ObjectId)ci.next();
        Individual indiv = doc.getIndividual(oid);
        if(indiv==null)
          continue;
        ObjectId scid = indiv.getSourceClasses()[0];
        if(scid!=null){
          String prefix = getPrefix(scid.getNameSpace());
          Element dim = document.createElement(prefix+":"+scid.getLocalName());
          String name = applyBase(oid.getId());
          dim.setAttribute("rdf:about",name);
          dism.appendChild(dim);
        }

      }
      diff.appendChild(dism);
      root.appendChild(diff);
    }

    //nameSpaces
    Iterator nsi = ns.keySet().iterator();
    while(nsi.hasNext()){
      String uri = (String)nsi.next();
      String pre = (String)ns.get(uri);
      if(pre.equals(""))
          root.setAttribute("xmlns",uri);
      else
          root.setAttribute("xmlns:"+pre,uri);
    }

    printOut();
  }


  /**
   * Create a element for an OntClass.
   */
  private Element writeEntity(Ontology ont, AxiomNode sub){
    Element eObject = null;
    Vector<Axiom> axioms;
    if(sub.getJoinType() == Ontology.ATOMICNODE){
    	axioms = (Vector<Axiom>)ont.getAxiomBySubject(sub.getAtomicNode());
    	OntObject ontObj = ont.getOntObject(sub.getAtomicNode());
    	if(ontObj!=null){
    		if(ontObj instanceof OntClass){
    			OntClass ontClass = (OntClass)ontObj;
    			String cname = ontClass.getID().getLocalName();
    			// 	comment&label
    			String comment = ontClass.getDescription();
    			String label = ontClass.getLabel();
    			if(ontClass.isDeprecated()){
    				eObject = document.createElement("owl:DeprecatedClass");
    				if(cname!=null&&!cname.equals(""))
    					eObject.setAttribute("rdf:ID",cname);
    			}
    			else{
    				eObject = document.createElement("owl:Class");
    				if(cname!=null&&!cname.equals(""))
    					eObject.setAttribute("rdf:ID",cname);
    			}

    			if(comment!=null){
    				Element ecomment = document.createElement("rdfs:comment");
    				Text text = document.createTextNode(comment);
    				ecomment.appendChild(text);
    				eObject.appendChild(ecomment);
    			}
    			if(label!=null){
    				Element elabel = document.createElement("rdfs:label");
    				Text text = document.createTextNode(label);
    				elabel.appendChild(text);
    				eObject.appendChild(elabel);
    			}
    		}else if(ontObj instanceof OntProperty){
    			OntProperty ontProperty = (OntProperty)ontObj;
    			if(ontProperty.isDatatype()!=null&&ontProperty.isDatatype().booleanValue())
    				eObject = document.createElement("owl:DatatypeProperty");
                if(ontProperty.isDatatype()!=null&&!ontProperty.isDatatype().booleanValue())
                	eObject = document.createElement("owl:ObjectProperty");
                if(ontProperty.isDatatype()==null)
                	eObject = document.createElement("rdf:Property");
                if(ontProperty.isAnnotation())
                	eObject = document.createElement("owl:AnnotationProperty");

                String name = ontProperty.getID().getLocalName();
                eObject.setAttribute("rdf:ID",name);
                if(ontProperty.isDeprecated()){
                  Element eptype = document.createElement("rdf:type");
                  eptype.setAttribute("rdf:resource", OWL.DeprecatedProperty.getURI());
                  eObject.appendChild(eptype);
                }
                if(ontProperty.isFunctional()){
                  Element eptype = document.createElement("rdf:type");
                  eptype.setAttribute("rdf:resource", OWL.FunctionalProperty.getURI());
                  eObject.appendChild(eptype);
                }

                if(ontProperty.isInverseFunctional()){
                  Element eptype = document.createElement("rdf:type");
                  eptype.setAttribute("rdf:resource", OWL.InverseFunctionalProperty.getURI());
                  eObject.appendChild(eptype);
                }

                if(ontProperty.isSymmetric()){
                  Element eptype = document.createElement("rdf:type");
                  eptype.setAttribute("rdf:resource", OWL.SymmetricProperty.getURI());
                  eObject.appendChild(eptype);
                }

                ObjectId domain = ontProperty.getDomain();
                if(domain!=null){
                  Element edomain = document.createElement("rdfs:domain");
                  name = applyBase(domain.getId());
                  edomain.setAttribute("rdf:resource",name);
                  eObject.appendChild(edomain);
                }

                ObjectId range = ontProperty.getRange();
                if(range!=null){
                  Element erange = document.createElement("rdfs:range");
                  name = applyBase(range.getId());
                  erange.setAttribute("rdf:resource",name);
                  eObject.appendChild(erange);
                }

                if(ontProperty.isTransitive()){
                  Element eptype = document.createElement("rdf:type");
                  eptype.setAttribute("rdf:resource", OWL.TransitiveProperty.getURI());
                  eObject.appendChild(eptype);
                }

                //comment&label
                String comment = ontProperty.getDescription();
                String label = ontProperty.getLabel();
                if(comment!=null){
                  Element ecomment = document.createElement("rdfs:comment");
                  Text text = document.createTextNode(comment);
                  ecomment.appendChild(text);
                  eObject.appendChild(ecomment);
                }
                if(label!=null){
                  Element elabel = document.createElement("rdfs:label");
                  Text text = document.createTextNode(label);
                  elabel.appendChild(text);
                  eObject.appendChild(elabel);
                }
    		}
        }else{
        	eObject = document.createElement("owl:Class");
        }

    	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
        	Axiom ax = it.next();
        	Element comp = handleAxiomRight(ont, ax);
        	eObject.appendChild(comp);
        }
    }else if(sub.getRestrictionId()!=null){
    	eObject = handleAxiomNodeRight(ont, sub);
    	axioms = new Vector<Axiom>();
    	Vector<Axiom> temp = (Vector<Axiom>)ont.getAllAxioms();
    	for(Iterator<Axiom> it = temp.iterator();it.hasNext();){
    		Axiom axiom = it.next();
    		if(axiom.getLeftAxiomNode().equals(sub)){
    			Element r = handleAxiomRight(ont, axiom);
    			eObject.appendChild(r);
    		}
    	}
    }else if(sub.getJoinType()>0){
      //zhp2 on 7/2/08: BUG: What if node is a collection? currently return null!
    }

    return eObject;
  }

  public Element handleAxiomRight(Ontology ont, Axiom axiom){
	  Element er = null;
	  boolean prop = false;
	  if(axiom.getType() == Ontology.COMPLEMENTOF){
		  er = document.createElement("owl:complementOf");
		  er.setAttribute("rdf:parseType","Collection");
	  }else if(axiom.getType() == Ontology.DISJOINTWITH)
		  er = document.createElement("owl:disjointWith");
	  else if(axiom.getType() == Ontology.EQUIVALENTCLASS)
		  er = document.createElement("owl:equivalentClass");
	  else if(axiom.getType() == Ontology.EQUIVALENTPROPERTY){
		  er = document.createElement("owl:equivalentProperty");
		  prop = true;
	  }else if(axiom.getType() == Ontology.INVERSEOF){
		  er = document.createElement("owl:inverseOf");
		  prop = true;
	  }else if(axiom.getType() == Ontology.SUBCLASSOF)
		  er = document.createElement("rdfs:subClassOf");
	  else if(axiom.getType() == Ontology.SUBPROPERTYOF){
		  er = document.createElement("rdfs:subPropertyOf");
		  prop = true;
	  }

	  AxiomNode right = axiom.getRightAxiomNode();
	  if(right.getRestrictionId()!=null){
		  Element r = handleAxiomNodeRight(ont, right);
		  er.appendChild(r);
	  }else{
		  if(right.getJoinType() == Ontology.ATOMICNODE){
			  String name = applyBase(right.getAtomicNode().getId());
	          er.setAttribute("rdf:resource",name);
		  }else if(right.getJoinType() > Ontology.ATOMICNODE){
			  Element equc;
			  if(prop)
				  equc = document.createElement("owl:ObjectProperty");
			  else
				  equc = document.createElement("owl:Class");

			  Element r = handleAxiomNodeRight(ont, right);
			  if(axiom.getType() == Ontology.EQUIVALENTCLASS && right.getJoinType()==Ontology.INTERSECTIONOF ||
					  right.getJoinType()==Ontology.UNIONOF || right.getJoinType()==Ontology.ONEOF)
				  return r;
			  equc.appendChild(r);
			  er.appendChild(equc);
		  }
	  }
	  return er;
  }

  public Element handleAxiomNodeRight(Ontology ont, AxiomNode right){
	  Element er = null;
	  boolean prop = false, oneOf = false;
	  if(right.getJoinType() > Ontology.ATOMICNODE){
		  if(right.getJoinType() == Ontology.INTERSECTIONOF){
			  er = document.createElement("owl:intersectionOf");
		  	  er.setAttribute("rdf:parseType","Collection");
		  }else if(right.getJoinType() == Ontology.UNIONOF){
			  er = document.createElement("owl:unionOf");
			  er.setAttribute("rdf:parseType","Collection");
		  }else if(right.getJoinType() == Ontology.ONEOF){
			  er = document.createElement("owl:oneOf");
			  er.setAttribute("rdf:parseType","Collection");
			  oneOf = true;
		  }else if(right.getJoinType() == Ontology.EQUIVALENTPROPERTY){
			  er = document.createElement("owl:equivalentProperty");
			  prop = true;
		  }else if(right.getJoinType() == Ontology.INVERSEOF){
			  er = document.createElement("owl:inverseOf");
			  prop = true;
		  }else if(right.getJoinType() == Ontology.SUBPROPERTYOF){
			  er = document.createElement("owl:subPropertyOf");
			  prop = true;
		  }

		  Vector<AxiomNode> children = right.getChildren();
		  for(Iterator<AxiomNode> it=children.iterator();it.hasNext();){
			  AxiomNode child = it.next();
			  if(child.getRestrictionId()!=null){
				  Element r = handleAxiomNodeRight(ont, child);
				  er.appendChild(r);
			  }else{
				  if(prop){
					  if(child.getJoinType() == Ontology.ATOMICNODE){
						  String name = applyBase(child.getAtomicNode().getId());
						  er.setAttribute("rdf:resource",name);
					  }
				  }else{
					  Element equc;
					  if(!oneOf)
						  equc = document.createElement("owl:Class");
					  else
						  equc = document.createElement("owl:Thing");

					  if(child.getJoinType() == Ontology.ATOMICNODE){
						  String name = applyBase(child.getAtomicNode().getId());
						  equc.setAttribute("rdf:about",name);
					  }else if(child.getJoinType() > Ontology.ATOMICNODE){
						  Element r = handleAxiomNodeRight(ont, child);
						  equc.appendChild(r);
					  }
					  er.appendChild(equc);
				  }
			 }
		 }
	  }else if(right.getJoinType() < Ontology.ATOMICNODE){
		  er = document.createElement("owl:Restriction");

		  Element onProperty = document.createElement("owl:onProperty");
		  AxiomNode property = right.getOnProperty();
		  if(property.getJoinType() == Ontology.ATOMICNODE){
			  String name = applyBase(property.getAtomicNode().getId());
			  onProperty.setAttribute("rdf:resource",name);
		  }else if(property.getJoinType() > Ontology.ATOMICNODE){
			  Element objProperty = document.createElement("owl:ObjectProperty");
			  Element r = handleAxiomNodeRight(ont, property);
			  objProperty.appendChild(r);
			  onProperty.appendChild(objProperty);
		  }
		  er.appendChild(onProperty);

		  AxiomNode allValsFrom = right.getAllValuesFrom();
		  if(allValsFrom!=null){
			  Element avf = document.createElement("owl:allValuesFrom");
			  if(allValsFrom.getJoinType() == Ontology.ATOMICNODE){
				  String name = applyBase(allValsFrom.getAtomicNode().getId());
				  avf.setAttribute("rdf:resource",name);
			  }else if(allValsFrom.getRestrictionId()!=null){
				  Element r = handleAxiomNodeRight(ont, allValsFrom);
				  avf.appendChild(r);
			  }else if(allValsFrom.getJoinType() > Ontology.ATOMICNODE){
				  Element allVals = document.createElement("owl:Class");
				  Element r = handleAxiomNodeRight(ont, allValsFrom);
				  allVals.appendChild(r);
				  avf.appendChild(allVals);
			  }
			  er.appendChild(avf);
		  }

		  AxiomNode someValsFrom = right.getSomeValuesFrom();
		  if(someValsFrom!=null){
			  Element svf = document.createElement("owl:someValuesFrom");
			  if(someValsFrom.getJoinType() == Ontology.ATOMICNODE){
				  String name = applyBase(someValsFrom.getAtomicNode().getId());
				  svf.setAttribute("rdf:resource",name);
			  }else if(someValsFrom.getRestrictionId()!=null){
				  Element r = handleAxiomNodeRight(ont, someValsFrom);
				  svf.appendChild(r);
			  }else if(someValsFrom.getJoinType() > Ontology.ATOMICNODE){
				  Element someVals = document.createElement("owl:Class");
				  Element r = handleAxiomNodeRight(ont, someValsFrom);
				  someVals.appendChild(r);
				  svf.appendChild(someVals);
			  }
			  er.appendChild(svf);
		  }

		  AxiomNode hasVal = right.getHasValue();
		  if(hasVal!=null){
			  Element hv = document.createElement("owl:hasValue");
			  if(hasVal.getJoinType() == Ontology.ATOMICNODE){
				  if(hasVal.getAtomicNode() instanceof DatatypeValue){
					  DatatypeValue dv = (DatatypeValue)hasVal.getAtomicNode();
					  hv.setAttribute("rdf:datatype",dv.getType());
				      hv.setNodeValue(dv.getValue());
				  }else{
					  String name = applyBase(hasVal.getAtomicNode().getId());
					  hv.setAttribute("rdf:resource",name);
				  }
			  }else if(hasVal.getRestrictionId()!=null){
				  Element r = handleAxiomNodeRight(ont, hasVal);
				  hv.appendChild(r);
			  }else if(hasVal.getJoinType() > Ontology.ATOMICNODE){
				  Element hasValue = document.createElement("owl:Class");
				  Element r = handleAxiomNodeRight(ont, hasVal);
				  hasValue.appendChild(r);
				  hv.appendChild(hasValue);
			  }
			  er.appendChild(hv);
		  }

		  // cardinality
		  int maxcard = right.getMaxCardinality();
		  int mincard = right.getMinCardinality();

		  if(mincard==maxcard){
		      Element ecard = document.createElement("owl:cardinality");
		      ecard.setAttribute("rdf:datatype","http://www.w3.org/2001/XMLSchema#nonNegativeInteger");
		      Text text = document.createTextNode(String.valueOf(mincard));
		      ecard.appendChild(text);
		      er.appendChild(ecard);
		  }

		  //maxCardinality
		  if(maxcard!=PropertyRestriction.INFINITY && maxcard!=mincard){
		      Element emaxcard = document.createElement("owl:maxCardinality");
		      emaxcard.setAttribute("rdf:datatype","http://www.w3.org/2001/XMLSchema#nonNegativeInteger");
		      Text text = document.createTextNode(String.valueOf(maxcard));
		      emaxcard.appendChild(text);
		      er.appendChild(emaxcard);
		  }

		  //minCardinality
		  if(mincard!=0&&mincard!=maxcard){
		      Element emincard = document.createElement("owl:minCardinality");
		      emincard.setAttribute("rdf:datatype","http://www.w3.org/2001/XMLSchema#nonNegativeInteger");
		      Text text = document.createTextNode(String.valueOf(mincard));
		      emincard.appendChild(text);
		      er.appendChild(emincard);
		  }
	  }
	  return er;
  }

  /**
   * Create an element for an individual.
   */
  private Element writeIndividual(edu.lehigh.swat.hawk.core.Document doc, Individual indiv){
    ObjectId[] source = indiv.getSourceClasses();
    int slen = source.length;
    String prefix = getPrefix(source[0].getNameSpace());
    String sourceName = "";
    if(prefix.equals(""))
      sourceName = source[0].getLocalName();
    else
      sourceName = prefix+":"+source[0].getLocalName();
    Element eindiv = document.createElement(sourceName);
    ObjectId iid = indiv.getID();
    String name = applyBase(iid.getId());
    if(!name.equals(""))
      eindiv.setAttribute("rdf:about",name);

    if(slen>1){
      for(int i=1;i<slen;i++){
        Element etype = document.createElement("rdf:type");
        etype.setAttribute("rdf:resource",source[i].getId());
        eindiv.appendChild(etype);
      }
    }
    //comment&label
    String comment = indiv.getDescription();
    String label = indiv.getLabel();
    if(comment!=null){
      Element ecomment = document.createElement("rdfs:comment");
      Text text = document.createTextNode(comment);
      ecomment.appendChild(text);
      eindiv.appendChild(ecomment);
    }
    if(label!=null){
      Element elabel = document.createElement("rdfs:label");
      Text text = document.createTextNode(label);
      elabel.appendChild(text);
      eindiv.appendChild(elabel);
    }
    //sameAs&differentFrom
    ObjectId[] sames = indiv.getAllSameAs();
    if(sames!=null){
      int len = sames.length;
      for(int i=0;i<len;i++){
        ObjectId oid = sames[i];
        Element esame = document.createElement("owl:sameIndividualAs");
        esame.setAttribute("rdf:resource",applyBase(oid.getId()));
        eindiv.appendChild(esame);
      }
    }

    ObjectId[] diffs = indiv.getAllDifferentFrom();
    if(diffs!=null){
      int len = diffs.length;
      for(int i=0;i<len;i++){
        ObjectId oid = diffs[i];
        Element ediff = document.createElement("owl:differentFrom");
        ediff.setAttribute("rdf:resource",applyBase(oid.getId()));
        eindiv.appendChild(ediff);
      }
    }

    ObjectId[] pids = indiv.getProperties();
    if(pids!=null){
      int len = pids.length;
      for(int i=0;i<len;i++){
        ObjectId pid = pids[i];
        ObjectId[] values = indiv.getValuesForProperty(pid);
        if(values!=null){
          int vlen = values.length;
          prefix = getPrefix(pid.getNameSpace());
          String pName = "";
          if(prefix.equals(""))
            pName = pid.getLocalName();
          else
            pName = prefix+":"+pid.getLocalName();
          for(int j=0;j<vlen;j++){
            Element epv = document.createElement(pName);
            if(values[j] instanceof DatatypeValue){
              DatatypeValue dv = (DatatypeValue)values[j];
              if(dv.getType()!=null)
                epv.setAttribute("rdf:datatype",dv.getType());
              Text text = document.createTextNode(dv.getValue());
              epv.appendChild(text);
            }else{
              name = applyBase(values[j].getId());
              epv.setAttribute("rdf:resource", name);
            }
            eindiv.appendChild(epv);

          }
        }
      }
    }
    return eindiv;
  }

  /**
   * Print out the DOM tree.
   */
  private void printOut(){
    try{
    OutputFormat format = new OutputFormat(document);
    format.setIndent(2);
    format.setLineSeparator(System.getProperty("line.separator"));
    format.setLineWidth(80);

    //format.setPreserveSpace(true);
    XMLSerializer serializer = new XMLSerializer (
        output, format);
    serializer.asDOMSerializer();
    serializer.serialize(document);
    }
    catch(Exception e){
      System.err.println(e.getMessage());
    }
  }

  /**
   * Get the prefix for nameSpace.
   */
  private String getPrefix(String uri){
    if(ns.containsKey(uri))
      return (String)ns.get(uri);
    else{
      String pre = "ns"+nsCount;
      nsCount++;
      ns.put(uri,pre);
      return pre;
    }
  }

  /**
   * Apply xml:base to shorten uri.
   */
  private String applyBase(String uri){
    if(uri==null||uri.equals(""))
      return uri;
    int p = uri.indexOf('#');
    if(p>0){
      String base = uri.substring(0,p);
      if(base.equalsIgnoreCase(xmlBase))
        return uri.substring(p,uri.length());
    }
    return uri;
  }

  private OutputStream output;
  private org.w3c.dom.Document document;
  private String xmlBase;
  private Map ns;
  private int nsCount = 0;


}
