/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/


package edu.lehigh.swat.hawk.storage.simpledb;

import edu.lehigh.swat.hawk.core.DocId;
import edu.lehigh.swat.hawk.core.ObjectId;
//import edu.lehigh.swat.hawk.storage.ontper.dldb.Implication;
import edu.lehigh.swat.hawk.core.Axiom;
import edu.lehigh.swat.hawk.core.AxiomNode;

import java.sql.*;
import java.util.*;

/**
 * <p>encapsulating the web ontology </p>
 * @author Z. Pan
 * @version 1.0
 */

public class Ontology extends Document implements edu.lehigh.swat.hawk.core.Ontology{

  /**
   * Constructor
   */
  public Ontology(edu.lehigh.swat.hawk.storage.simplememory.Ontology mOnto) {
    memoryOnto = mOnto;
  }

  /******************** Axiom Design Methods ***************************/

  public void addAxiom(Axiom a){
	  memoryOnto.addAxiom(a);
  }

  public void addAxiomByObject(ObjectId obj, Axiom a){
	  memoryOnto.addAxiomByObject(obj, a);
  }

  public void addAxiomBySubject(ObjectId sub, Axiom a){
	  memoryOnto.addAxiomBySubject(sub, a);
  }

  public void addAxiomByType(Integer type, Axiom a){
	  memoryOnto.addAxiomByType(type, a);
  }

  public void addAxiomNodeRestriction(AxiomNode res){
	  memoryOnto.addAxiomNodeRestriction(res);
  }

  public void addAllAxiomNodeRestrictions(Vector<AxiomNode> res){
	  memoryOnto.addAllAxiomNodeRestrictions(res);
  }

  public List<Axiom> getAllAxioms(){
	  return memoryOnto.getAllAxioms();
  }

  public Hashtable<ObjectId, Vector<Axiom>> getAllAxiomsByObject(){
	  return memoryOnto.getAllAxiomsByObject();
  }

  public Hashtable<ObjectId, Vector<Axiom>> getAllAxiomsBySubject(){
	  return memoryOnto.getAllAxiomsBySubject();
  }

  public Hashtable<Integer, Vector<Axiom>> getAllAxiomsByType(){
	  return memoryOnto.getAllAxiomsByType();
  }

  public List<Axiom> getAxiomByObject(ObjectId object){
	  return memoryOnto.getAxiomByObject(object);
  }

  public List<Axiom> getAxiomBySubject(ObjectId subject){
	  return memoryOnto.getAxiomBySubject(subject);
  }

  public List<Axiom> getAxiomByType(Integer type){
	  return memoryOnto.getAxiomByType(type);
  }

  public Vector<AxiomNode> getAxiomNodeRestrictions(){
	  return memoryOnto.getAxiomNodeRestrictions();
  }
  /*********************************************************************/

  /**
   * Get the memory-based ontology object.
   */
  public edu.lehigh.swat.hawk.storage.simplememory.Ontology getMemOnto(){
    return memoryOnto;
  }

  /**
   * get the ID of this document
   */
  public DocId getID(){
    return memoryOnto.getID();
  }

  /**
   * set the ID of this document
   */
  public void setID(DocId id){
    memoryOnto.setID(id);
  }

  /**
   * Add an imports statement.
   */
  public void addImports(DocId id){
    super.addImports(id);
    memoryOnto.addImports(id);
    Document idoc = (Document)model.getDocument(id.getId());
    if(!(idoc instanceof Ontology))
      return;
    Ontology ionto = (Ontology)idoc;
    edu.lehigh.swat.hawk.core.OntClass[] ics = ionto.getAllPlusImportedClasses();
    int len = ics.length;
    for(int i=0;i<len;i++){
      edu.lehigh.swat.hawk.core.OntClass icls = ics[i];
 //     edu.lehigh.swat.hawk.core.OntClass oc = createClass();
 //     oc.setID(icls.getID());
      addImportedClass(icls.getID(),icls);
    }
    edu.lehigh.swat.hawk.core.OntProperty[] ips = ionto.getAllPlusImportedProperties();
    int plen = ips.length;
    for(int i=0;i<plen;i++){
      edu.lehigh.swat.hawk.core.OntProperty ip = ips[i];
//      edu.lehigh.swat.hawk.core.OntProperty op = createProperty();
//      op.setID(ip.getID());
      addImportedProperty(ip.getID(),ip);
    }
  }


  /**
   * add a class to this ontology
   */
  public void addOntClass(edu.lehigh.swat.hawk.core.OntClass c){
    memoryOnto.addOntClass(c);
    String tableName = model.getTableName(c.getID());
    if(tableName==null)
      return;
    String qs = "CREATE TABLE "+tableName
    +" ( ID INTEGER not null, SOURCE INTEGER not null,ONTO INTEGER, PRIMARY KEY(ID,SOURCE) )";

    try{
      stmt.executeUpdate(qs);
    }
     catch(SQLException e){
                System.out.println("error when store class");
                System.out.println(qs);
                System.out.println("SQLState:"+e.getSQLState());
                System.out.println("Message :"+e.getMessage());
                System.out.println("Vendor  :"+e.getErrorCode());
     }

      String qs2 = "";
      String viewName = tableName+"_view";

      qs2 = "CREATE VIEW "+viewName+" AS\n SELECT * FROM "+tableName;

      String qs3 = "";

      String qs4= "Insert into tables_index values ('"+tableName+"', "+ontSeq+", "+memoryOnto.getClassIndex(c.getID())+")"; //zhp2: insert tables_index
      try{
        qs3 = "INSERT INTO view_definition VALUES ('"+viewName+"', '"+qs2+"')";
        stmt.executeUpdate(qs2);
        stmt.executeUpdate(qs3);
        stmt.executeUpdate(qs4);
      }
      catch(SQLException e){
        System.out.println("error when store class, create view");
        System.out.println(qs4);
        System.out.println("SQLState:"+e.getSQLState());
        System.out.println("Message :"+e.getMessage());
        System.out.println("Vendor  :"+e.getErrorCode());
      }
  }

  /**
   *
   * get the index number for a class
   */
  public int getClassIndex(ObjectId cid){
      return memoryOnto.getClassIndex(cid);
  }

  /**
   * get the specified class
   */
  public edu.lehigh.swat.hawk.core.OntClass getOntClass(ObjectId id){
    edu.lehigh.swat.hawk.core.OntClass oc = memoryOnto.getOntClass(id);
    return oc;
  }

  /**
   * get the object (class/property) given the ID
   */
  public edu.lehigh.swat.hawk.core.OntObject getOntObject(ObjectId id){
    edu.lehigh.swat.hawk.core.OntObject obj = memoryOnto.getOntObject(id);
    return obj;
  }

  /**
   * delete the specified class
   */
  public void deleteOntClass(ObjectId id){
    memoryOnto.deleteOntClass(id);
  }

  /**
   * get all the classes in this ontology
   */
  public edu.lehigh.swat.hawk.core.OntClass[] getAllClasses(){
    return memoryOnto.getAllClasses();

  }

  /**
   * Get All the classes plus the imported Classes.
   */
  public edu.lehigh.swat.hawk.core.OntClass[] getAllPlusImportedClasses(){
    return memoryOnto.getAllPlusImportedClasses();
  }


  /**
   * Get the Iterator of all the classes.
   */
  public Iterator getClassIterator(){
    return memoryOnto.getClassIterator();
  }


  /**
   * add a property to this ontology
   */
  public void addProperty(edu.lehigh.swat.hawk.core.OntProperty prop){
    memoryOnto.addProperty(prop);
    if(dbName==null)
        dbName = ((DatabaseStorage)getModel()).getDBName();
    String tableName = model.getTableName(prop.getID());
    if(tableName==null)
      return;
    String qs = "";
    String pinfo = "INSERT INTO propertytypes VALUES('";
    tableName = tableName+"_p";
    if((prop.isDatatype()!=null && !prop.isDatatype().booleanValue())||
        prop.isFunctional()||prop.isTransitive()||prop.isInverseFunctional()||
         prop.isSymmetric()){
      qs = "CREATE TABLE "+tableName
    +" ( SUBJECT INTEGER not null, OBJECT INTEGER not null, SOURCE INTEGER not null, ONTO INTEGER,"+
    " PRIMARY KEY(SUBJECT,OBJECT,SOURCE), unique(OBJECT))";
      pinfo += tableName+"', 'object')";
    }
    else{
      ObjectId range = prop.getRange();
      if(range!=null&&range.getNameSpace().equalsIgnoreCase("http://www.w3.org/2001/XMLSchema#")){
          if(range.getLocalName().toLowerCase().endsWith("integer")){
              qs = "CREATE TABLE "+tableName
                   +" ( SUBJECT INTEGER not null, OBJECT INTEGER not null,"+
                   " SOURCE INTEGER not null, ONTO INTEGER,"+
                   " PRIMARY KEY(SUBJECT,OBJECT,SOURCE))";
              pinfo += tableName+"', 'int')";
          }
          else if(range.getLocalName().toLowerCase().endsWith("float")){
              qs = "CREATE TABLE "+tableName
                   +" (SUBJECT INTEGER not null, OBJECT FLOAT not null,"+
                   " SOURCE INTEGER not null, ONTO INTEGER,"+
                   " PRIMARY KEY(SUBJECT,OBJECT,SOURCE))";
              pinfo += tableName+"', 'float')";
          }
          else if(range.getLocalName().toLowerCase().endsWith("date")){ // Ameet Aug 30, 2007
              qs = "CREATE TABLE "+tableName
                   +" (SUBJECT INTEGER not null, OBJECT DATE not null,"+
                   " SOURCE INTEGER not null, ONTO INTEGER,"+
                   " PRIMARY KEY(SUBJECT,OBJECT,SOURCE))";
              pinfo += tableName+"', 'date')";
          }
      }
      if(qs.length()<1){
          if(dbName.indexOf(".mdb")>0)
              qs = "CREATE TABLE " + tableName
               + " (SUBJECT INTEGER, OBJECT MEMO,"
               +" SOURCE INTEGER,ONTO INTEGER," +
               " INDEX(SUBJECT), INDEX(OBJECT))";
          else if(dbName.indexOf("jdbc:db2")>0){
              qs = "CREATE TABLE " + tableName
                + " (SUBJECT INTEGER not null, OBJECT CLOB(4096),"
                +" SOURCE INTEGER not null, ONTO INTEGER," +
                " unique (SUBJECT))";
          }
          else
             qs = "CREATE TABLE " + tableName
               + " (SUBJECT INTEGER, OBJECT TEXT,"
               +" SOURCE INTEGER, ONTO INTEGER," +
               " INDEX (SUBJECT), INDEX(OBJECT(15)))";
          pinfo += tableName + "', 'datatype')";
      }
    }
   // log.println(qs);

    try{
      stmt.executeUpdate(qs);
      stmt.executeUpdate(pinfo);
    }
    catch(SQLException e){
                //log.println("error when store property");
      System.out.println("error when store property");
      System.out.println("SQLState:"+e.getSQLState());
      System.out.println("Message :"+e.getMessage());
      System.out.println("Vendor  :"+e.getErrorCode());
    }

      String qs2 = "";
      String viewName = tableName+"_view";
  //    if(viewSeqNum == currentSeqNum){

      qs2 = "CREATE VIEW "+viewName+" AS\n SELECT * FROM "+tableName;

      String qs3 = "";

      String qs4= "Insert into tables_index values ('"+tableName+"', "+ontSeq+", "+memoryOnto.getPropertyIndex(prop.getID())+")"; //zhp2: insert tables_index
      try{
        qs3 = "INSERT INTO view_definition VALUES ('"+viewName+"', '"+qs2+"')";
        stmt.executeUpdate(qs2);
        stmt.executeUpdate(qs3);
        stmt.executeUpdate(qs4);
      }
      catch(SQLException e){
              //  log.println("error when store class, create view");
        System.out.println(qs4);
        System.out.println("error when store property, create view");
        System.out.println("SQLState:"+e.getSQLState());
        System.out.println("Message :"+e.getMessage());
        System.out.println("Vendor  :"+e.getErrorCode());
      }
  }

  /**
   *
   * get the index number for a property
   */
  public int getPropertyIndex(ObjectId pid){
      return memoryOnto.getPropertyIndex(pid);
  }

  /**
   * get the specified property
   */
  public edu.lehigh.swat.hawk.core.OntProperty getProperty(ObjectId id){
    return memoryOnto.getProperty(id);
  }

  /**
   * delete a specified property
   */
  public void deleteProperty(ObjectId id){
    memoryOnto.deleteProperty(id);
  }

  /**
   * get all the properties in this ontology
   */
  public edu.lehigh.swat.hawk.core.OntProperty[] getAllProperties(){
    return memoryOnto.getAllProperties();
  }

  /**
   * Get All the classes plus the imported properties.
   */
  public edu.lehigh.swat.hawk.core.OntProperty[] getAllPlusImportedProperties(){
    return memoryOnto.getAllPlusImportedProperties();
  }

  /**
   * Get the Iterator of all the properties.
   */
  public Iterator getPropertyIterator(){
    return memoryOnto.getPropertyIterator();
  }


  /**
   * add a property to this ontology
   */
  public void addRestriction(edu.lehigh.swat.hawk.core.PropertyRestriction p){
    memoryOnto.addRestriction(p);
  }

  /**
   * get the specified property
   */
  public edu.lehigh.swat.hawk.core.PropertyRestriction getRestriction(ObjectId id){
    return memoryOnto.getRestriction(id);
  }

  /**
   * delete a specified property
   */
  public void deleteRestriction(ObjectId id){
    memoryOnto.deleteRestriction(id);
  }

  /**
   * get all the properties in this ontology
   */
  public edu.lehigh.swat.hawk.core.PropertyRestriction[] getAllRestrictions(){
    return memoryOnto.getAllRestrictions();
  }

  /**
   * Get the priorVersion ontologies which are backwardCompatibleWith
   */
  public DocId[] getCompatiblePriorOnts(){
    DocId[] dIds = memoryOnto.getCompatiblePriorOnts();
    return dIds;
  }

  /**
   * Add a priorVersion ontology which is backwardCompatibleWith.
   */
  public void addCompatiblePriorOnts(DocId id){
    memoryOnto.addCompatiblePriorOnts(id);
  }

  /**
   * Delete a priorVersion ontology which is backwardCompatibleWith.
   */
  public void deleteCompatiblePriorOnts(DocId id){
    memoryOnto.deleteCompatiblePriorOnts(id);
  }

  /**
   * Get the priorVersion ontologies may be incompatibleWith
   */
  public DocId[] getPriorOnts(){
    return memoryOnto.getPriorOnts();
  }

  /**
   * Add a priorVersion ontology may be incompatibleWith.
   */
  public void addPriorOnts(DocId id){
    memoryOnto.addPriorOnts(id);
  }

  /**
   * Delete a priorVersion ontology may be incompatibleWith.
   */
  public void deletePriorOnts(DocId id){
    memoryOnto.deletePriorOnts(id);
  }

  /**
   * get the version of this ontology
   */
  public String getVersion(){
    return memoryOnto.getVersion();
  }

  /**
   * Set the version of this ontology.
   */
  public void setVersion(String ver){
    memoryOnto.setVersion(ver);
  }


  /**
   * Create a Class object.
   */
  public edu.lehigh.swat.hawk.core.OntClass createClass(){
    return memoryOnto.createClass();
  }

  /**
   * Create a Property object.
   */
  public edu.lehigh.swat.hawk.core.OntProperty createProperty(){
    return memoryOnto.createProperty();
  }

  /**
   * Create a PropertyRestriction object.
   */
  public edu.lehigh.swat.hawk.core.PropertyRestriction createRestriction(){
    return memoryOnto.createRestriction();
  }

  /**
   * Flush the axioms into database.
   */
  public void flush(){
      try{
          super.flush();
          if (dbName == null)
              dbName = ((DatabaseStorage) getModel()).getDBName();
          flushSubs();
          flushEquivals(); //z.pan on 02/23/07: may already been taken care of by subs
          flushInverseOf();
//    flushIntersections();
          if (getAllProperties().length > 0) {
              model.updatePropertyTypes(); //zhop2: every time add a new ontology, update the property types info.
          }
      }catch(Exception e){
          System.out.println(e.getMessage());
          e.printStackTrace();
      }
    try{
          model.storeOntology(this);
      }catch(Exception e){
          System.out.println(e.getMessage());
          e.printStackTrace();
      }
  }

  protected void flushSubs(){
    //deal with subClass  //pzx: changed to all plus imports on 10/12/2004
    edu.lehigh.swat.hawk.core.OntClass[] classes = this.getAllPlusImportedClasses();
    int cNum = classes.length;

    for(int i=0;i<cNum;i++){
      edu.lehigh.swat.hawk.core.OntClass cls = classes[i];
      if(cls.getID().getLocalName()==null||cls.getID().getLocalName()=="")
        continue;
      if(importedClass(cls.getID())&&!assertedClass(cls.getID()))
         continue;

      ObjectId[] sid = cls.getDirectSubClasses();
      String className = model.getTableName(cls.getID());
  //    int oriSeq = model.getOntSeq(cls.getID().getNameSpace());
      int sn =0;
      if(sid!=null)
        sn = sid.length;
      if(sn<=0)
        continue;
      String vName = className+"_view";
      String dv = "DROP TABLE "+vName;
      StringBuffer cvb;

      String def = model.getViewDefinition(vName);
      if(def==null){
          System.out.println("view definition not found:" + vName);
          continue;
      }
      else
        cvb = new StringBuffer(def);

      for(int j=0;j<sn;j++){
   //     int sOriSeq = model.getOntSeq(sid[j].getNameSpace());
   //     if(sOriSeq!=oriSeq)
   //       continue;

        String tName = model.getTableName(sid[j]);
        if(tName==null)
          continue;

        String avName = tName+"_view";

        if(cvb.indexOf(" "+avName+" ")>0)
         continue;

        cvb.append(" UNION ALL SELECT * FROM ");
        String vdef = model.getViewDefinition(avName);
        if(vdef!=null){
          int ind = vdef.indexOf(" FROM");
          vdef = vdef.substring(ind+5);
        }
        if(vdef!=null&&vdef.indexOf(" "+className)>0){
          int i1 = vdef.indexOf(" "+className);
          int i2 = vdef.indexOf(' ',i1+1);
          if(i2<=0)
            i2 = vdef.length();
          int i0 = i1-24; //was i1-20
          StringBuffer sb = new StringBuffer(vdef);
          sb.delete(i0,i2);
          if(sb.indexOf(" "+className)>0){
              i1 = sb.indexOf(" "+className);
              i2 = sb.indexOf(" UNION ", i1);
              if(i2<=0)
                  i2=sb.length();
              sb.delete(i1-24,i2); //zhp2: was i1-20
         }
          cvb.append(sb);
        }
        else
          cvb.append(tName).append("_view");
      }
      try{
        String cv = cvb.toString();
        if(dbName.endsWith(".mdb")|| dbName.indexOf("h2")>0){
          stmt.execute(dv);
          stmt.execute(cv);
        }else if(dbName.indexOf("sql")>0&&dbName.indexOf("hsql")<0)
          stmt.execute(cvb.insert(7,"OR REPLACE ").toString());

        String qs3 = "UPDATE view_definition SET definition='"+cv+"'"+
             "WHERE viewname='"+vName+"'";
        stmt.execute(qs3);
      }
      catch(SQLException e){
        System.out.println("error when recreate class hierarchy");
        System.out.println(cvb.toString());
        System.out.println("SQLState:"+e.getSQLState());
        System.out.println("Message :"+e.getMessage());
        System.out.println("Vendor  :"+e.getErrorCode());
      }
    }

   // deal with subProperty
   edu.lehigh.swat.hawk.core.OntProperty[] Properties
       = this.getAllPlusImportedProperties();
   int pNum = Properties.length;

   for(int i=0;i<pNum;i++){
     edu.lehigh.swat.hawk.core.OntProperty cls = Properties[i];
     if(cls.getID().getLocalName()==null||cls.getID().getLocalName()=="")
       continue;
     if(importedProperty(cls.getID())&&!assertedProperty(cls.getID()))
         continue;

     ObjectId[] sid = cls.getDirectSubProperties();
     String propertyName = model.getTableName(cls.getID())+"_p";
   //  int oriSeq = model.getOntSeq(cls.getID().getNameSpace());
     if(sid==null)
       continue;
     int sn = sid.length;
     if(sn<=0)
       continue;
     String vName = propertyName+"_view";
     String dv = "DROP TABLE "+vName;
     StringBuffer cvb;

     String def = model.getViewDefinition(vName);
     if(def==null){
         System.out.println("view definition not found:" + vName);
         continue;
     }
     else
     cvb = new StringBuffer(def);

     for(int j=0;j<sn;j++){
    //    int sOriSeq = model.getOntSeq(sid[j].getNameSpace());
   //   if(sOriSeq!=oriSeq)
  //        continue;
       String tName = model.getTableName(sid[j]);
       if(tName==null)
         continue;
       tName = tName + "_p";
       String avName = tName+"_view";

       if(cvb.indexOf(" "+avName+" ")>0)
         continue;

       cvb.append(" UNION ALL SELECT * FROM ");
       String vdef = model.getViewDefinition(avName);
       if(vdef!=null){
         int ind = vdef.indexOf(" FROM");
         vdef = vdef.substring(ind+5);
       }
       if(vdef!=null&&vdef.indexOf(" "+propertyName)>0){
         int i1 = vdef.indexOf(" "+propertyName);
         int i2 = vdef.indexOf(' ',i1+1);
         if(i2<=0)
           i2 = vdef.length();
         int i0 = i1-24;   //zhp2:was i1-20
         StringBuffer sb = new StringBuffer(vdef);
         sb.delete(i0,i2);
         if(sb.indexOf(" "+propertyName)>0){
             i1 = sb.indexOf(" "+propertyName);
             i2 = sb.indexOf(" UNION ", i1);
             if(i2<=0)
                 i2=sb.length();
             sb.delete(i1-24,i2);  //was i1-20
         }
         cvb.append(sb);
       }
       else
         cvb.append(tName).append("_view");
     }
     try{
       String cv = cvb.toString();
       if(dbName.endsWith(".mdb")|| dbName.indexOf("h2")>0){
         stmt.execute(dv);
         stmt.execute(cv);
       }else if(dbName.indexOf("sql")>0&&dbName.indexOf("hsql")<0)
         stmt.execute(cvb.insert(7,"OR REPLACE ").toString());

       String qs3 = "UPDATE view_definition SET definition='"+cv+"'"+
            "WHERE viewname='"+vName+"'";
       stmt.execute(qs3);

     }
     catch(SQLException e){
       System.out.println("error when recreate property hierarchy");
       System.out.println(cvb.toString());
       System.out.println("SQLState:"+e.getSQLState());
       System.out.println("Message :"+e.getMessage());
       System.out.println("Vendor  :"+e.getErrorCode());
     }
   }
  }

  protected void flushEquivals(){
    //deal with equivalentClass
    edu.lehigh.swat.hawk.core.OntClass[] classes = this.getAllClasses();
    int cNum = classes.length;
    for(int i=0;i<cNum;i++){
      edu.lehigh.swat.hawk.core.OntClass cls = classes[i];
      if(cls.getID().getLocalName()==null||cls.getID().getLocalName()=="")
        continue;
      ObjectId[] eid = cls.getEquivalentClasses();
      String className = model.getTableName(cls.getID());
    //  int oriSeq = model.getOntSeq(cls.getID().getNameSpace());
      if(eid==null)
        continue;
      int sn = eid.length;
      if(sn<=0)
        continue;
      String vName = className+"_view";
      String dv = "DROP TABLE "+vName;
      StringBuffer cvb;
      String start = model.getViewDefinition(vName);
      cvb = new StringBuffer(start);

      for(int j=0;j<sn;j++){
     //   int sOriSeq = model.getOntSeq(eid[j].getNameSpace());
      //  if(sOriSeq!=oriSeq)
      //    continue;

        String tName = model.getTableName(eid[j]);
        if(tName == null)
          continue;
        cvb.append(" UNION ALL SELECT * FROM ");
        String avName = tName+"_view";
        String vdef = model.getViewDefinition(avName);
        if(vdef!=null){
          int ind = vdef.indexOf(" FROM");
          vdef = vdef.substring(ind+5);
        }
        cvb.append(vdef);
      }
      try{
        String cv = cvb.toString();
        if(dbName.endsWith(".mdb")|| dbName.indexOf("h2")>0){
         stmt.execute(dv);
         stmt.execute(cv);
       }else if(dbName.indexOf("sql")>0&&dbName.indexOf("hsql")<0)
         stmt.execute(cvb.insert(7,"OR REPLACE ").toString());

        String qs3 = "UPDATE view_definition SET definition='"+cv+"'"+
            "WHERE viewname='"+vName+"'";
        stmt.execute(qs3);

      }
      catch(SQLException e){
        System.out.println("error when recreate class hierarchy");
        System.out.println(cvb.toString());
        System.out.println("SQLState:"+e.getSQLState());
        System.out.println("Message :"+e.getMessage());
        System.out.println("Vendor  :"+e.getErrorCode());
      }
    }

    //deal with equivalentProperty
    edu.lehigh.swat.hawk.core.OntProperty[] Properties = this.getAllProperties();
    int pNum = Properties.length;
    for(int i=0;i<pNum;i++){
      edu.lehigh.swat.hawk.core.OntProperty cls = Properties[i];
      if(cls.getID().getLocalName()==null||cls.getID().getLocalName()=="")
        continue;
      ObjectId[] eid = cls.getEquivalentProperties();
      String PropertyName = model.getTableName(cls.getID())+"_p";
  //    int oriSeq = model.getOntSeq(cls.getID().getNameSpace());
      if(eid==null)
        continue;
      int sn = eid.length;
      if(sn<=0)
        continue;
      String vName = PropertyName+"_view";
      String dv = "DROP TABLE "+vName;
      StringBuffer cvb;
      String start = model.getViewDefinition(vName);
      cvb = new StringBuffer(start);

      for(int j=0;j<sn;j++){
    //    int sOriSeq = model.getOntSeq(eid[j].getNameSpace());
     //   if(sOriSeq!=oriSeq)
     //     continue;

        String tName = model.getTableName(eid[j]);
        if(tName == null)
          continue;
        tName = tName + "_p";
        cvb.append(" UNION ALL SELECT * FROM ");
        String avName = tName+"_view";
        String vdef = model.getViewDefinition(avName);
        if(vdef!=null){
          int ind = vdef.indexOf(" FROM");
          vdef = vdef.substring(ind+5);
        }
        cvb.append(vdef);
      }
      try{
        String cv = cvb.toString();
        if(dbName.endsWith(".mdb")|| dbName.indexOf("h2")>0){
         stmt.execute(dv);
         stmt.execute(cv);
       }else if(dbName.indexOf("sql")>0&&dbName.indexOf("hsql")<0)
         stmt.execute(cvb.insert(7,"OR REPLACE ").toString());

        String qs3 = "UPDATE view_definition SET definition='"+cv+"'"+
            "WHERE viewname='"+vName+"'";
        stmt.execute(qs3);

      }
      catch(SQLException e){
        System.out.println("error when recreate Property hierarchy");
        System.out.println(cvb.toString());
        System.out.println("SQLState:"+e.getSQLState());
        System.out.println("Message :"+e.getMessage());
        System.out.println("Vendor  :"+e.getErrorCode());
      }
    }
  }

  /**
   * Flush owl:inverseOf into db view.
   */
  protected void flushInverseOf(){
    edu.lehigh.swat.hawk.core.OntProperty[] Properties
        = this.getAllPlusImportedProperties();
    int pNum = Properties.length;

    for (int i = 0; i < pNum; i++) {
      edu.lehigh.swat.hawk.core.OntProperty cls = Properties[i];
      if (cls.getID().getLocalName() == null || cls.getID().getLocalName() == "")
        continue;
     if(!assertedProperty(cls.getID()))
         continue;

      ObjectId[] sid = cls.getInverseOf();
      String propertyName = model.getTableName(cls.getID())+"_p";
  //  int oriSeq = model.getOntSeq(cls.getID().getNameSpace());
      if (sid == null)
        continue;
      int sn = sid.length;
      if (sn <= 0)
        continue;
      String vName = propertyName  + "_inv"+ "_view";
      String dv = "DROP TABLE " + vName;
      StringBuffer cvb;

      String def = model.getViewDefinition(vName);
//      if(def==null)
        cvb = new StringBuffer("CREATE VIEW " + vName + " AS SELECT * FROM "
                               + propertyName + "_view ");
//      else
//        cvb = new StringBuffer(def);

      for (int j = 0; j < sn; j++) {
  //    int sOriSeq = model.getOntSeq(sid[j].getNameSpace());
  //   if(sOriSeq!=oriSeq)
  //        continue;
        String tName = model.getTableName(sid[j]);
        if (tName == null)
          continue;
        tName = tName + "_p_view";
        cvb.append(" UNION ALL SELECT ");
        cvb.append(tName + ".OBJECT AS SUBJECT, ").append(tName
              + ".SUBJECT AS OBJECT, "+tName+".source, "+tName
              +".onto "+"FROM ").append(tName);
      }
      try {
        String cv = cvb.toString();
        if (dbName.endsWith(".mdb")|| dbName.indexOf("h2")>0) {
          if(def!=null)
            stmt.execute(dv);
          stmt.execute(cv);
        }
        else if (dbName.indexOf("sql") > 0 && dbName.indexOf("hsql")<0)
          stmt.execute(cvb.insert(7, "OR REPLACE ").toString());

        String qs3 = "UPDATE view_definition SET definition='" + cv + "'" +
            "WHERE viewname='" + vName + "'";
        stmt.execute(qs3);
        String qs4 = "UPDATE propertytypes SET type='hasinverse' WHERE pname='"
            +propertyName+"'";
        stmt.execute(qs4);
      }
      catch (SQLException e) {
        System.out.println("error when recreate property hierarchy");
        System.out.println(cvb.toString());
        System.out.println("SQLState:" + e.getSQLState());
        System.out.println("Message :" + e.getMessage());
        System.out.println("Vendor  :" + e.getErrorCode());
      }
    }
  }

  private void flushIntersections(){
    //deal with equivalentClass
    edu.lehigh.swat.hawk.core.OntClass[] classes = this.getAllClasses();
    int cNum = classes.length;
    for(int i=0;i<cNum;i++){
      edu.lehigh.swat.hawk.core.OntClass cls = classes[i];
      if(cls.getID().getLocalName()==null||cls.getID().getLocalName()=="")
        continue;
      ObjectId[] eid = cls.getIntersectionOfs();
      String className = model.getTableName(cls.getID());
    //  int oriSeq = model.getOntSeq(cls.getID().getNameSpace());
      if(eid==null)
        continue;
      int sn = eid.length;
      if(sn<=0)
        continue;
      String vName = className+"_int_view";
      String dv = "DROP TABLE "+vName;
      StringBuffer cvb = new StringBuffer("CREATE VIEW " + vName + " AS ");
      String start = model.getViewDefinition(vName);
    //  cvb = new StringBuffer(start);

      String procName = className+"_proc";
      StringBuffer cpb = new StringBuffer("CREATE PROCEDURE "+procName+" () \nBEGIN\nflush tables;\n");

      for(int j=0;j<sn;j++){
     //   int sOriSeq = model.getOntSeq(eid[j].getNameSpace());
      //  if(sOriSeq!=oriSeq)
      //    continue;

        String tName = model.getTableName(eid[j]);
        String ivName = tName;
        if(tName != null){
            ivName = tName + "_view";
            cpb.append("drop temporary table if exists "+ivName+";\n");
            cpb.append("CREATE temporary table "+ivName+" (id int, source int, onto int, key(id)) select * from "+ivName+";\n");
            cpb.append("flush tables;\n");
        }
        else{
            String[] ret = this.ViewForRestriction(eid[j]);
            if(ret==null){
                cvb=null;
                break;
            }

            ivName = ret[0];
            cpb.append(ret[1]);
        }
        if(ivName == null)
          continue;
        if(j==0){
            if(ivName.startsWith("(SELECT"))
                cvb.append(ivName.substring(1)).deleteCharAt(cvb.length()).append(" where ID IN ()");
            else
                cvb.append("SELECT * FROM ").append(ivName + " where ID IN ()");
        }
        else{
            int fromind = ivName.indexOf(" FROM ");
            int commaind = ivName.indexOf(",");
            if(fromind<=0)
                cvb.insert(cvb.length() - j,
                       "SELECT ID FROM " + ivName + " AND ID IN()");
            else
                cvb.insert(cvb.length()-j,
                ivName.substring(1,commaind)+ivName.substring(fromind)+" AND ID IN()");
        }
      }
      if(cvb==null)
          continue;
      cvb.delete(cvb.length()-13, cvb.length()-1);
      cpb.append("drop temporary table if exists "+vName+";\n");
      cpb.append("CREATE temporary table "+vName+" (id int, source int, onto int, key(id)) "+cvb.substring(cvb.indexOf(" AS ")+4)+");\n");
      cpb.append("flush tables;\n");
      cpb.append("END");
      try{
        String cv = cvb.toString();
        if(dbName.endsWith(".mdb")|| dbName.indexOf("h2")>0){
            if(start!=null)
                stmt.execute(dv);
         stmt.execute(cv);
       }else if(dbName.indexOf("sql")>0){
//           stmt.execute(cvb.insert(7, "OR REPLACE ").toString());
           String cp = cpb.toString();
           stmt.execute(cp);
       }
        String qs3 = "UPDATE view_definition SET definition='"+cv+"'"+
            "WHERE viewname='"+vName+"'";
        stmt.execute(qs3);

      }
      catch(SQLException e){
        System.out.println("error when flush intersections");
        System.out.println(cvb.toString());
        System.out.println(cpb.toString());
        System.out.println("SQLState:"+e.getSQLState());
        System.out.println("Message :"+e.getMessage());
        System.out.println("Vendor  :"+e.getErrorCode());
      }
    }
  }


  /**
   * Add an imported class.
   */
  public void addImportedClass(ObjectId cid, edu.lehigh.swat.hawk.core.OntClass c){
    memoryOnto.addImportedClass(cid,c);
  }

  /**
   * Add an imported property.
   */
  public void addImportedProperty(ObjectId pid, edu.lehigh.swat.hawk.core.OntProperty p){
    memoryOnto.addImportedProperty(pid,p);
  }

  /**
   * Add an individual.
   */
  public void addIndividual(edu.lehigh.swat.hawk.core.Individual indiv){
    super.addIndividual(indiv);
    memoryOnto.addIndividual(indiv);
  }

  /**
   * To see if ontology has the class.
   */
  public boolean hasClass(String name){
    return memoryOnto.hasClass(name);
  }

  /**
   * Get the imported property.
   */
  public boolean hasProperty(String name){
    return memoryOnto.hasProperty(name);
  }

  /**
   * To tell if the ontology asserted specified class.
   */
  public boolean assertedClass(ObjectId cid){
      return memoryOnto.assertedClass(cid);
  }

  /**
   * To tell if the ontology asserted specified property.
   */
  public boolean assertedProperty(ObjectId pid){
      return memoryOnto.assertedProperty(pid);
  }

  /**
   * Add an asserted class.
   */
  public void addAssertedClass(ObjectId cid){
      memoryOnto.addAssertedClass(cid);
  }

  /**
   * Add an asserted property.
   */
  public void addAssertedProperty(ObjectId pid){
      memoryOnto.addAssertedProperty(pid);
  }

  /**
   * Get the table names in the view definition
   */
  protected Vector viewDefToTabNames(String vdef){
    StringTokenizer st = new StringTokenizer(vdef);
    Vector v = new Vector();
    String t;
    while(st.hasMoreTokens()){
      t = st.nextToken();
      if(t.equalsIgnoreCase("from")){
        String viewName = st.nextToken();
        int ind2 = viewName.lastIndexOf("_view");
        if(ind2>0){
          viewName = viewName.substring(0,ind2);
          int ind1 = viewName.lastIndexOf('_');
          viewName = viewName.substring(0,ind1);
        }
        v.add(viewName);
      }
    }
    return v;
  }

  /**
   * Set the document sequence number.
   */
  public void setOntSeq(int i){
    ontSeq = i;
  }

  /**
   * Get the document sequence number.
   */
  public int getOntSeq(){
    return ontSeq;
  }

  /**
   * To tell if the ontology imorted specified class.
   */
  public boolean importedClass(ObjectId cid){
    return memoryOnto.importedClass(cid);
  }

  /**
   * To tell if the ontology imported specified property.
   */
  public boolean importedProperty(ObjectId pid){
    return memoryOnto.importedProperty(pid);
  }

  /**
   * Get view definition for restriction.
   * @param id ObjectId
   * @return String[]
   * the first string is the view definition, the second one is for MySQL only.
   */
  public String[] ViewForRestriction(ObjectId id){
      edu.lehigh.swat.hawk.storage.simplememory.PropertyRestriction res =
       (edu.lehigh.swat.hawk.storage.simplememory.PropertyRestriction)memoryOnto.getRestriction(id);
      ObjectId prop = res.getOnProperty();
      ObjectId svf = res.getSomeValuesFrom();
      if(svf==null)
          return null;
      String pname = model.getTableName(prop)+"_p_view";
      String cname = model.getTableName(svf)+"_view";
//      String tname = id.getId().replace('[',' ').replace(']',' ').replace('-','_').trim().substring(22);
      String qs = "(SELECT subject as ID, "+pname+".source as source, "+
                  pname+".onto as onto"+" FROM "+pname+", "+cname+" where "
                  +cname+".id="+pname+".object)";
      String tmp1 = "drop temporary table if exists "+pname+";\n"
                    +"CREATE temporary table "+pname+" (subject int, object int, source int, onto int, key(subject), key(object)) select * from "+pname+";\n";
      String tmp2 = "drop temporary table if exists "+cname+";\n"
                    +"CREATE temporary table "+cname+" (id int, source int, onto int, key(id)) select * from "+cname+";\n";
      String resTName = id.getId().replace(']',' ').trim().substring(id.getId().lastIndexOf(':')+2);
      String tmp3 = "drop temporary table if exists "+resTName+";\n"
                    +"CREATE temporary table "+resTName
                    +" (id int, source int, onto int, key(id)) select "+pname+".subject as id, "
                    +pname+".source as source, "+pname+".onto as onto"
                    +" from "+pname+", "+cname+" where "
                    +cname+".id="+pname+".object;\n";
      String tmp = tmp1+tmp2+tmp3;
      if(dbName.indexOf(".mdb")>0 || dbName.indexOf("h2")>0)
          return new String[]{qs,tmp};
      return new String[]{resTName,tmp};
  }
  /**
   * Get the imported class.
   */
//  public edu.lehigh.swat.hawk.core.OntClass getImportedClass(ObjectId id){
//    return memoryOnto.getImportedClass(id);
//  }

  /**
   * To see if ontology has imported the property.
   */
//  public edu.lehigh.swat.hawk.core.OntProperty getImportedProperty(ObjectId id){
//    return memoryOnto.getImportedProperty(id);
//  }

  protected int ontSeq;
  protected edu.lehigh.swat.hawk.storage.simplememory.Ontology memoryOnto;
//  private Vector compatibleOntos;
  protected String dbName;

}
