/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/

package edu.lehigh.swat.hawk.storage.simplememory;
import edu.lehigh.swat.hawk.core.Axiom;
import edu.lehigh.swat.hawk.core.AxiomNode;
import edu.lehigh.swat.hawk.core.ObjectId;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;

/**
 * A simple implementation of core.OntClass in memory.
 * @author Z. Pan
 * @version 1.0
 */

public class OntClass extends OntObject
    implements edu.lehigh.swat.hawk.core.OntClass{


  /**
   * Constructor
   */
  public OntClass() {
    subClasses = new Vector();
    superClasses = new Vector();
    complementOf = new Vector();
    unionOf = new Vector();
    oneOf = new Vector();
    intersectionOf = new Vector();
    equivalentClasses = new Vector();
    propertyRestrictions = new Vector();
    disjointWith = new Vector();
    deprecated = false;
    satisfiable = null;
  }

  /**
   * Add a direct subClass.
   * $memo: currently we only set bi-directional links to those classes who are
   * in the same document.
   */
  public void addSubClass(ObjectId c){
    /*if(!subClasses.contains(c)){
      subClasses.add(c);
      edu.lehigh.swat.hawk.core.Ontology ont = this.getOntology();
      edu.lehigh.swat.hawk.core.OntClass cc = ont.getOntClass(c);
      if(cc!=null&&!cc.equals(this))
        cc.addSuperClass(this.getID());
    }*/

    Ontology ont = (Ontology)this.getOntology();
    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.SUBCLASSOF));
    if(axioms!=null){
    	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
    		Axiom temp = it.next();
    		if(temp.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getLeftAxiomNode().getAtomicNode().equals(c) &&
    				temp.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getRightAxiomNode().getAtomicNode().equals(this.getID()))
    			return;
    	}
    }

    Axiom axiom = new Axiom(Ontology.SUBCLASSOF);
	AxiomNode left = new AxiomNode();
	left.setAtomicNode(c);
	AxiomNode right = new AxiomNode();
	right.setAtomicNode(this.getID());
	axiom.setLeftAxiomNode(left);
	axiom.setRightAxiomNode(right);

	ont.addAxiom(axiom);
	ont.addAxiomBySubject(c, axiom);
	ont.addAxiomByObject(this.getID(), axiom);
	ont.addAxiomByType(new Integer(Ontology.SUBCLASSOF), axiom);
  }

  /**
   * get the direct subClasses
   */
  public ObjectId[] getDirectSubClasses(){
    /*int len = subClasses.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = subClasses.toArray();
    System.arraycopy(oa,0,od,0,len);*/
	ArrayList<ObjectId> od = new ArrayList<ObjectId>();
    Ontology ont = (Ontology)this.getOntology();
    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.SUBCLASSOF));
    if(axioms!=null){
    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
    	Axiom a = it.next();
    	if(a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(this.getID()) &&
    			a.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE){
    		od.add(a.getLeftAxiomNode().getAtomicNode());
    	}
    }
    }
    ObjectId[] temp = new ObjectId[od.size()];
    System.arraycopy(od.toArray(),0,temp,0,temp.length);
	return temp;
  }

  /**
   * Delete a direct subClass.
   */
  public void deleteSubClass(ObjectId c){
    /*if(subClasses.contains(c)){
      subClasses.remove(c);
      edu.lehigh.swat.hawk.core.Ontology ont = this.getOntology();
      edu.lehigh.swat.hawk.core.OntClass cc = ont.getOntClass(c);
      if (cc != null)
        cc.deleteSuperClass(this.getID());
    }*/
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.SUBCLASSOF));
	if(axioms!=null){
    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
    	Axiom a = it.next();
    	if(a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(this.getID()) &&
    			a.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getLeftAxiomNode().getAtomicNode().equals(c)){
    		ont.getAllAxioms().remove(a);
    		ont.getAxiomBySubject(c).remove(a);
    		ont.getAxiomByObject(this.getID()).remove(a);
    		ont.getAxiomByType(new Integer(Ontology.SUBCLASSOF)).remove(a);
    		break;
    	}
    }
	}
  }

  /**
   * Add a direct super Class.
   */
  public void addSuperClass(ObjectId c){
    /*if(!superClasses.contains(c)){
      superClasses.add(c);
      edu.lehigh.swat.hawk.core.Ontology ont = this.getOntology();
      edu.lehigh.swat.hawk.core.OntClass cc = ont.getOntClass(c);
      if(cc!=null&&!cc.equals(this))
        cc.addSubClass(this.getID());
    }*/
	    Ontology ont = (Ontology)this.getOntology();
	    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.SUBCLASSOF));
	    if(axioms!=null){
	    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	    	Axiom temp = it.next();
	    	if(temp.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getLeftAxiomNode().getAtomicNode().equals(this.getID()) &&
	    	   temp.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getRightAxiomNode().getAtomicNode().equals(c))
	    		return;
	    }
	    }
	    Axiom axiom = new Axiom(Ontology.SUBCLASSOF);
		AxiomNode left = new AxiomNode();
		left.setAtomicNode(this.getID());
		AxiomNode right = new AxiomNode();
		right.setAtomicNode(c);
		axiom.setLeftAxiomNode(left);
		axiom.setRightAxiomNode(right);

		ont.addAxiom(axiom);
		ont.addAxiomBySubject(this.getID(), axiom);
		ont.addAxiomByObject(c, axiom);
		ont.addAxiomByType(new Integer(Ontology.SUBCLASSOF), axiom);

  }

  /**
   * get the direct super Classes
   */
  public ObjectId[] getDirectSuperClasses(){
    /*int len = superClasses.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = superClasses.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;*/
	ArrayList<ObjectId> od = new ArrayList<ObjectId>();
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.SUBCLASSOF && a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE){
	   		od.add(a.getRightAxiomNode().getAtomicNode());
	   	}
	}
	}
	ObjectId[] temp = new ObjectId[od.size()];
    System.arraycopy(od.toArray(),0,temp,0,temp.length);
	return temp;
  }

  /**
   * delete a direct super Class
   */
  public void deleteSuperClass(ObjectId c){
    /*if(superClasses.contains(c)){
      superClasses.remove(c);
      edu.lehigh.swat.hawk.core.Ontology ont = this.getOntology();
      edu.lehigh.swat.hawk.core.OntClass cc = ont.getOntClass(c);
      if (cc != null)
        cc.deleteSubClass(this.getID());
    }*/
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.SUBCLASSOF &&
			a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(c)){
	  		ont.getAllAxioms().remove(a);
	   		ont.getAxiomBySubject(this.getID()).remove(a);
	   		ont.getAxiomByObject(c).remove(a);
	   		ont.getAxiomByType(new Integer(Ontology.SUBCLASSOF)).remove(a);
	   		break;
	   	}
	}
	}
  }

  /**
   * Add a property restriction.
   */
  public void addPropertyRestriction(ObjectId r){
    propertyRestrictions.add(r);
  }

  /**
   * Delete a property restriction.
   */
  public void deletePropertyRestriction(ObjectId r){
    propertyRestrictions.remove(r);
  }

  /**
   * Get the property restrictions.
   */
  public ObjectId[] getPropertyRestrictions(){
    int len = propertyRestrictions.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = propertyRestrictions.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;
  }

  /**
   * Add a equivalent class.
   */
  public void addEquivalentClass(ObjectId ec){
      if(ec.equals(this.getID()))
        return;

	    Ontology ont = (Ontology)this.getOntology();
	    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.EQUIVALENTCLASS));
	    if(axioms!=null){
	    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	    	Axiom temp = it.next();
	    	if(temp.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE &&
	    	   ((temp.getLeftAxiomNode().getAtomicNode().equals(this.getID()) && temp.getRightAxiomNode().getAtomicNode().equals(ec)) ||
	    	    (temp.getLeftAxiomNode().getAtomicNode().equals(ec) && temp.getRightAxiomNode().getAtomicNode().equals(this.getID()))) )
	    		return;
	    }
	    }
	    Axiom axiom = new Axiom(Ontology.EQUIVALENTCLASS);
		AxiomNode left = new AxiomNode();
		left.setAtomicNode(this.getID());
		AxiomNode right = new AxiomNode();
		right.setAtomicNode(ec);
		axiom.setLeftAxiomNode(left);
		axiom.setRightAxiomNode(right);

		ont.addAxiom(axiom);
		ont.addAxiomBySubject(this.getID(), axiom);
		ont.addAxiomByObject(ec, axiom);
		ont.addAxiomByType(new Integer(Ontology.EQUIVALENTCLASS), axiom);
  }

  /**
   * Delete a equivalent class.
   */
  public void deleteEquivalentClass(ObjectId ec){
    //equivalentClasses.remove(ec);
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.EQUIVALENTCLASS &&
			a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(ec)){
	  		ont.getAllAxioms().remove(a);
	   		ont.getAxiomBySubject(this.getID()).remove(a);
	   		ont.getAxiomByObject(ec).remove(a);
	   		ont.getAxiomByType(new Integer(Ontology.EQUIVALENTCLASS)).remove(a);
	   		return;
	   	}
	}
	}
	axioms = (Vector<Axiom>)ont.getAxiomByObject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.EQUIVALENTCLASS &&
			a.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getLeftAxiomNode().getAtomicNode().equals(ec)){
	  		ont.getAllAxioms().remove(a);
	   		ont.getAxiomBySubject(ec).remove(a);
	   		ont.getAxiomByObject(this.getID()).remove(a);
	   		ont.getAxiomByType(new Integer(Ontology.EQUIVALENTCLASS)).remove(a);
	   		return;
	   	}
	}
	}
  }

  /**
   * Get the equivalent classes.
   */
  public ObjectId[] getEquivalentClasses(){
    /*int len = equivalentClasses.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = equivalentClasses.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;*/
	ArrayList<ObjectId> od = new ArrayList<ObjectId>();

	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.EQUIVALENTCLASS && a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE)
	   			od.add(a.getRightAxiomNode().getAtomicNode());
	}
	}
	axioms = (Vector<Axiom>)ont.getAxiomByObject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.EQUIVALENTCLASS &&	a.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE)
	   			od.add(a.getLeftAxiomNode().getAtomicNode());
	}
	}
	ObjectId[] temp = new ObjectId[od.size()];
    System.arraycopy(od.toArray(),0,temp,0,temp.length);
	return temp;
  }

  /**
   * Add an 'UnionOf' class.
   */
  public void addUnionOf(ObjectId u){
    //unionOf.add(u);

	  Ontology ont = (Ontology)this.getOntology();
	  Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	  if(axioms!=null){
	  for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
		  Axiom a = it.next();
		  if(a.getType()==Ontology.EQUIVALENTCLASS &&	a.getRightAxiomNode().getJoinType() == Ontology.UNIONOF){
			  AxiomNode node = new AxiomNode();
		   	  node.setAtomicNode(u);
		   	  a.getRightAxiomNode().addChild(node);
		   	  break;
		  }
	  }
	  }
  }

  /**
   * Delete an 'UnionOf' class.
   */
  public void deleteUnionOf(ObjectId u){
    //unionOf.remove(u);
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.EQUIVALENTCLASS &&	a.getRightAxiomNode().getJoinType() == Ontology.UNIONOF){
	   		for(Iterator<AxiomNode> itC = a.getRightAxiomNode().getChildren().iterator();itC.hasNext();){
	   			AxiomNode child = itC.next();
	   			if((child.getJoinType() == Ontology.ATOMICNODE && child.getAtomicNode().equals(u)) ||
	   			   (child.getRestrictionId()!=null && child.getRestrictionId().equals(u.getId()))){
	   				a.getRightAxiomNode().getChildren().remove(child);
	   				return;
	   			}
	   		}
	   	}
	}
	}
  }

  /**
   * Get the 'UnionOf' classes.
   */
  public ObjectId[] getUnionOfs(){
    /*int len = unionOf.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = unionOf.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;*/

	System.out.println("This method is no longer supported, try - AxiomNode[] getUnionAxiomNodes()");
	return null;

    // If we change the body of this method to use the Axiom/AxiomNode design then there are some issues. There could be an axiom like
    // A = (B intersection C) U (D intersection E), where this method may not be able to return anything. So the best thing would be to return
    // the actual right axiom node i.e. (B intersection C) U (D intersection E)
  }

  /**
   * This new method returns the right an array of axiomNodes each of which represents a unionOf entity.
   * @return
   */
  public AxiomNode[] getUnionAxiomNodes(){
	  ArrayList<AxiomNode> od = new ArrayList<AxiomNode>();
	  Ontology ont = (Ontology)this.getOntology();
	  Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	  if(axioms!=null){
	  for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
		  Axiom a = it.next();
		  if(a.getType()==Ontology.EQUIVALENTCLASS && a.getRightAxiomNode().getJoinType() == Ontology.UNIONOF)
		   	od.add(a.getRightAxiomNode());
	  }
	  }
	  	AxiomNode[] temp = new AxiomNode[od.size()];
	    System.arraycopy(od.toArray(),0,temp,0,temp.length);
		return temp;
  }

  /**
   * Add an 'IntersectionOf' class or restriction.
   */
  public void addIntersectionOf(ObjectId i){
    //intersectionOf.add(i);
	  Ontology ont = (Ontology)this.getOntology();
	  Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	  if(axioms!=null){
	  for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
		  Axiom a = it.next();
		  if(a.getType()==Ontology.EQUIVALENTCLASS &&	a.getRightAxiomNode().getJoinType() == Ontology.INTERSECTIONOF){
			  AxiomNode node = new AxiomNode();
		   	  node.setAtomicNode(i);
		   	  a.getRightAxiomNode().addChild(node);
		   	  break;
		  }
	  }
	  }
  }

  /**
   * Delete an 'IntersectionOf' class or restriction.
   */
  public void deleteIntersectionOf(ObjectId i){
    //intersectionOf.remove(i);
	  Ontology ont = (Ontology)this.getOntology();
	  Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	  if(axioms!=null){
	  for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
		  Axiom a = it.next();
		  if(a.getType()==Ontology.EQUIVALENTCLASS && a.getRightAxiomNode().getJoinType() == Ontology.INTERSECTIONOF){
		   	for(Iterator<AxiomNode> itC = a.getRightAxiomNode().getChildren().iterator();itC.hasNext();){
		   		AxiomNode child = itC.next();
		   		if((child.getJoinType() == Ontology.ATOMICNODE && child.getAtomicNode().equals(i)) ||
		   		   (child.getRestrictionId()!=null && child.getRestrictionId().equals(i.getId()))){
		   			a.getRightAxiomNode().getChildren().remove(child);
		   			return;
		   		}
		   	}
		 }
	  }
	 }
  }

  /**
   * Get the 'IntersectionOf' classes or restrictions.
   */
  public ObjectId[] getIntersectionOfs(){
    /*int len = intersectionOf.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = intersectionOf.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;*/
	System.out.println("This method is no longer supported, try - AxiomNode[] getIntersectionAxiomNodes()");
	return null;
    //  If we change the body of this method to use the Axiom/AxiomNode design then there are some issues. There could be an axiom like
    // A = (B U C) intersection (D U E), where this method may not be able to return anything. So the best thing would be to return
    // the actual right axiom node i.e. (B U C) intersection (D U E)
  }

  /**
   * This new method returns the right an array of axiomNodes each of which represents a intersectionOf entity.
   * @return
   */
  public AxiomNode[] getIntersectionAxiomNodes(){
	  ArrayList<AxiomNode> od = new ArrayList<AxiomNode>();
	  Ontology ont = (Ontology)this.getOntology();
	  Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	  if(axioms!=null){
	  for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
		  Axiom a = it.next();
		  if(a.getType()==Ontology.EQUIVALENTCLASS && a.getRightAxiomNode().getJoinType() == Ontology.INTERSECTIONOF)
		   	od.add(a.getRightAxiomNode());
	  }
	  }
	  	AxiomNode[] temp = new AxiomNode[od.size()];
	    System.arraycopy(od.toArray(),0,temp,0,temp.length);
		return temp;
  }

  /**
   * Add an 'OneOf' object.
   */
  public void addOneOf(ObjectId one){
    //oneOf.add(one);
	  	Ontology ont = (Ontology)this.getOntology();
	    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.ONEOF));
	    if(axioms!=null){
	    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	    	Axiom temp = it.next();
	    	if(temp.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getLeftAxiomNode().getAtomicNode().equals(this.getID()) &&
	    	   temp.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getRightAxiomNode().getAtomicNode().equals(one))
	    		return;
	    }
	    }
	    Axiom axiom = new Axiom(Ontology.ONEOF);
		AxiomNode left = new AxiomNode();
		left.setAtomicNode(this.getID());
		AxiomNode right = new AxiomNode();
		right.setAtomicNode(one);
		axiom.setLeftAxiomNode(left);
		axiom.setRightAxiomNode(right);

		ont.addAxiom(axiom);
		ont.addAxiomBySubject(this.getID(), axiom);
		ont.addAxiomByObject(one, axiom);
		ont.addAxiomByType(new Integer(Ontology.ONEOF), axiom);
  }

  /**
   * Delete an 'OneOf' object.
   */
  public void deleteOneOf(ObjectId one){
    //oneOf.remove(one);
	  Ontology ont = (Ontology)this.getOntology();
	  Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	  if(axioms!=null){
	  for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
		  Axiom a = it.next();
		  if(a.getType()==Ontology.EQUIVALENTCLASS && a.getRightAxiomNode().getJoinType() == Ontology.ONEOF){
		   	for(Iterator<AxiomNode> itC = a.getRightAxiomNode().getChildren().iterator();itC.hasNext();){
		   		AxiomNode child = itC.next();
		   		if((child.getJoinType() == Ontology.ATOMICNODE && child.getAtomicNode().equals(one)) ||
		   		   (child.getRestrictionId()!=null && child.getRestrictionId().equals(one.getId()))){
		   			a.getRightAxiomNode().getChildren().remove(child);
		   			return;
		   		}
		   	}
		 }
	  }
	 }
  }

  /**
   * Get the 'OneOf' objects.
   */
  public ObjectId[] getOneOfs(){
    int len = oneOf.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = oneOf.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;

    /*ArrayList<ObjectId> od = new ArrayList<ObjectId>();
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.ONEOF && a.getRightAxiomNode().getAtomicNode()!=null){
	   		od.add(a.getRightAxiomNode().getAtomicNode());
	   	}
	}
	return (ObjectId[])od.toArray();*/
  }

  /**
   * Add an 'ComplementOf' class or restriction.
   */
  public void addComplementOf(ObjectId i){
    //complementOf.add(i);
	  	Ontology ont = (Ontology)this.getOntology();
	    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.COMPLEMENTOF));
	    if(axioms!=null){
	    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	    	Axiom temp = it.next();
	    	if(temp.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getLeftAxiomNode().getAtomicNode().equals(this.getID()) &&
	    	   temp.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getRightAxiomNode().getAtomicNode().equals(i))
	    		return;
	    }
	    }
	    Axiom axiom = new Axiom(Ontology.COMPLEMENTOF);
		AxiomNode left = new AxiomNode();
		left.setAtomicNode(this.getID());
		AxiomNode right = new AxiomNode();
		right.setAtomicNode(i);
		axiom.setLeftAxiomNode(left);
		axiom.setRightAxiomNode(right);

		ont.addAxiom(axiom);
		ont.addAxiomBySubject(this.getID(), axiom);
		ont.addAxiomByObject(i, axiom);
		ont.addAxiomByType(new Integer(Ontology.COMPLEMENTOF), axiom);
  }

  /**
   * Delete an 'ComplementOf' class or restriction.
   */
  public void deleteComplementOf(ObjectId i){
    //complementOf.remove(i);
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.COMPLEMENTOF && a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(i)){
	  		ont.getAllAxioms().remove(a);
	   		ont.getAxiomBySubject(this.getID()).remove(a);
	   		ont.getAxiomByObject(i).remove(a);
	   		ont.getAxiomByType(new Integer(Ontology.COMPLEMENTOF)).remove(a);
	   		break;
	   	}
	 }
	}
  }

  /**
   * Get the 'ComplementOf' classes or restrictions.
   */
  public ObjectId[] getComplementOfs(){
    /*int len = complementOf.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = complementOf.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;*/

    ArrayList<ObjectId> od = new ArrayList<ObjectId>();
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.COMPLEMENTOF && a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE)
	   		od.add(a.getRightAxiomNode().getAtomicNode());
	}
	}
	ObjectId[] temp = new ObjectId[od.size()];
    System.arraycopy(od.toArray(),0,temp,0,temp.length);
	return temp;
  }

  /**
   * Add a 'disjointWith' class.
   */
  public void addDisjointWith(ObjectId id){
    //disjointWith.add(id);

	  	Ontology ont = (Ontology)this.getOntology();
	    Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomByType(new Integer(Ontology.DISJOINTWITH));
	    if(axioms!=null){
	    for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	    	Axiom temp = it.next();
	    	if(temp.getLeftAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getLeftAxiomNode().getAtomicNode().equals(this.getID()) &&
	    	   temp.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && temp.getRightAxiomNode().getAtomicNode().equals(id))
	    		return;
	    }
	    }

	    Axiom axiom = new Axiom(Ontology.DISJOINTWITH);
		AxiomNode left = new AxiomNode();
		left.setAtomicNode(this.getID());
		AxiomNode right = new AxiomNode();
		right.setAtomicNode(id);
		axiom.setLeftAxiomNode(left);
		axiom.setRightAxiomNode(right);

		ont.addAxiom(axiom);
		ont.addAxiomBySubject(this.getID(), axiom);
		ont.addAxiomByObject(id, axiom);
		ont.addAxiomByType(new Integer(Ontology.DISJOINTWITH), axiom);
  }

  /**
   * Delete a 'disjointWith' class.
   */
  public void deleteDisjointWith(ObjectId id){
    //disjointWith.remove(id);
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.DISJOINTWITH &&
	       a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE && a.getRightAxiomNode().getAtomicNode().equals(id)){
	  		ont.getAllAxioms().remove(a);
	   		ont.getAxiomBySubject(this.getID()).remove(a);
	   		ont.getAxiomByObject(id).remove(a);
	   		ont.getAxiomByType(new Integer(Ontology.DISJOINTWITH)).remove(a);
	   		break;
	   	}
	 }
	}
  }

  /**
   * Get the disjointWith classes.
   */
  public ObjectId[] getDisjointWith(){
    /*int len = disjointWith.size();
    ObjectId[] od = new ObjectId[len];
    Object[] oa = disjointWith.toArray();
    System.arraycopy(oa,0,od,0,len);
    return od;*/

    ArrayList<ObjectId> od = new ArrayList<ObjectId>();
	Ontology ont = (Ontology)this.getOntology();
	Vector<Axiom> axioms = (Vector<Axiom>)ont.getAxiomBySubject(this.getID());
	if(axioms!=null){
	for(Iterator<Axiom> it = axioms.iterator();it.hasNext();){
	  	Axiom a = it.next();
	   	if(a.getType()==Ontology.DISJOINTWITH && a.getRightAxiomNode().getJoinType() == Ontology.ATOMICNODE)
	   		od.add(a.getRightAxiomNode().getAtomicNode());
	}
	}

	ObjectId[] temp = new ObjectId[od.size()];
    System.arraycopy(od.toArray(),0,temp,0,temp.length);
	return temp;
  }

  /**
   * Test if this class is deprecated.
   */
  public boolean isDeprecated(){
    return deprecated;
  }

  /**
   * Set the class to be deprecated.
   */
  public void setDeprecated(boolean dep){
    deprecated = dep;
  }

  /**
   * Set the satisfiability of this class.
   */
  public void setSatisfiable(Boolean sat){
    satisfiable = sat;
  }

  /**
   * Test if this class is satisfiable.
   */
  public Boolean isSatisfiabele(){
    return satisfiable;
  }

  /**
   * Add descendant to this class.
   * @param sid ObjectId
   */
  public void addDescendant(ObjectId sid){
      if(!subClasses.contains(sid))
          subClasses.add(sid);
  }

  /**
   * Test if this class subsumes the given class.
   * @param sid ObjectId
   * @return boolean
   */
  public boolean subsumes(ObjectId sid){
      if(subClasses.contains(sid))
          return true;
      else
          return false;
  }
  private Boolean satisfiable;
  private boolean deprecated;
  private Vector subClasses, superClasses, complementOf, unionOf,
                 oneOf, intersectionOf, equivalentClasses, disjointWith,
                 propertyRestrictions;

}
