/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/


package edu.lehigh.swat.hawk.storage.simplememory;

import java.util.*;
import edu.lehigh.swat.hawk.core.ObjectId;
import edu.lehigh.swat.hawk.core.DocId;
import edu.lehigh.swat.hawk.core.Axiom;
import edu.lehigh.swat.hawk.core.AxiomNode;
/**
 * A simple implementation of core.ontology in memory.
 * @author Z. Pan
 * @version 1.0
 */

public class Ontology extends Document implements edu.lehigh.swat.hawk.core.Ontology{

    private static final long serialVersionUID = 1L;

  /**
   * Constructor
   */
  public Ontology() {
	axioms = new Vector<Axiom>();
	axiomsBySubject = new Hashtable<ObjectId, Vector<Axiom>>();
	axiomsByObject = new Hashtable<ObjectId, Vector<Axiom>>();
	axiomsByType = new Hashtable<Integer, Vector<Axiom>>();
	axiomNodeRestrictions = new Vector<AxiomNode>();
    classes = new Hashtable();
    properties = new Hashtable();
    restrictions = new Hashtable();
    compatiblePriorOnts = new Vector();
    priorOnts = new Vector();
    version = null;
    importedClasses = new Hashtable();
    importedProperties = new Hashtable();
    classIds = new Vector();
    propertyIds = new Vector();
    assertedClasses = new Vector();
    assertedProperties = new Vector();
  }

  public void addAxiom(Axiom a){
	  axioms.add(a);
  }

  public List<Axiom> getAllAxioms(){
	  return axioms;
  }

  public Hashtable<ObjectId, Vector<Axiom>> getAllAxiomsBySubject(){
	  return axiomsBySubject;
  }

  public Hashtable<Integer, Vector<Axiom>> getAllAxiomsByType(){
	  return axiomsByType;
  }

  public Hashtable<ObjectId, Vector<Axiom>> getAllAxiomsByObject(){
      return axiomsByObject;
  }

  public void addAxiomBySubject(ObjectId sub, Axiom a){
	  if(axiomsBySubject.containsKey(sub))
		  axiomsBySubject.get(sub).add(a);
	  else{
		  Vector<Axiom> axm = new Vector<Axiom>();
		  axm.add(a);
		  axiomsBySubject.put(sub, axm);
	  }
  }

  public void addAxiomByObject(ObjectId obj, Axiom a){
	  if(axiomsByObject.containsKey(obj))
		  axiomsByObject.get(obj).add(a);
	  else{
		  Vector<Axiom> axm = new Vector<Axiom>();
		  axm.add(a);
		  axiomsByObject.put(obj, axm);
	  }
  }

  public void addAxiomByType(Integer type, Axiom a){
	  if(axiomsByType.containsKey(type))
		  axiomsByType.get(type).add(a);
	  else{
		  Vector<Axiom> axm = new Vector<Axiom>();
		  axm.add(a);
		  axiomsByType.put(type, axm);
	  }
  }

  public List<Axiom> getAxiomBySubject(ObjectId subject){
	  if(subject == null)
		  return null;
	  if(axiomsBySubject.containsKey(subject))
		  return axiomsBySubject.get(subject);
	  return null;
 }

  public List<Axiom> getAxiomByObject(ObjectId object){
	  if(object == null)
		  return null;
	  if(axiomsByObject.containsKey(object))
		  return axiomsByObject.get(object);
	  return null;
  }

  public List<Axiom> getAxiomByType(Integer type){
	  if(axiomsByType.containsKey(type))
		  return axiomsByType.get(type);
	  return null;
  }

  public Vector<AxiomNode> getAxiomNodeRestrictions(){
	  return axiomNodeRestrictions;
  }

  public void addAxiomNodeRestriction(AxiomNode res){
	  axiomNodeRestrictions.add(res);
  }

  public void addAllAxiomNodeRestrictions(Vector<AxiomNode> res){
	  axiomNodeRestrictions.addAll(res);
  }

  /**
   * add a class to this ontology
   */
  public void addOntClass(edu.lehigh.swat.hawk.core.OntClass c){
    classes.put(c.getID(),c);
    classIds.add(c.getID());
  }

  /**
   *
   * get the index number for a class
   */
  public int getClassIndex(ObjectId cid){
      return classIds.indexOf(cid)+1;
  }

  /**
   * get the object (class/property) given the ID
   */
  public edu.lehigh.swat.hawk.core.OntObject getOntObject(ObjectId id){
    if(classes.containsKey(id))
    	return (edu.lehigh.swat.hawk.core.OntClass)classes.get(id);
    else if(importedClasses.containsKey(id))
    	return (edu.lehigh.swat.hawk.core.OntClass) importedClasses.get(id);
    else if(properties.containsKey(id))
    	return (edu.lehigh.swat.hawk.core.OntProperty)properties.get(id);
    else if(importedProperties.containsKey(id))
      return (edu.lehigh.swat.hawk.core.OntProperty)importedProperties.get(id);
    return null;
  }

  /**
   * get the specified class
   */
  public edu.lehigh.swat.hawk.core.OntClass getOntClass(ObjectId id){
    if(classes.containsKey(id))
      return (edu.lehigh.swat.hawk.core.OntClass)classes.get(id);
    if(importedClasses.containsKey(id)){
      return (edu.lehigh.swat.hawk.core.OntClass) importedClasses.get(id);
    }
    return null;
  }

  /**
   * delete the specified class
   */
  public void deleteOntClass(ObjectId id){
    if(classes.containsKey(id))
      classes.remove(id);
  }

  /**
   * get all the classes in this ontology
   */
  public edu.lehigh.swat.hawk.core.OntClass[] getAllClasses(){
    Collection c = classes.values();
    Object[] oa = c.toArray();
    int len = oa.length;
    edu.lehigh.swat.hawk.core.OntClass[] oc = new edu.lehigh.swat.hawk.core.OntClass[len];
    System.arraycopy(oa,0,oc,0,len);
    return oc;

  }

  /**
   * Get the Iterator of all the classes.
   */
  public Iterator getClassIterator(){
    return classes.values().iterator();
  }


  /**
   * add a property to this ontology
   */
  public void addProperty(edu.lehigh.swat.hawk.core.OntProperty p){
    properties.put(p.getID(),p);
    propertyIds.add(p.getID());
  }

  /**
   *
   * get the index number for a property
   */
  public int getPropertyIndex(ObjectId pid){
      int i = propertyIds.indexOf(pid)+1;
      String s = "-"+String.valueOf(i);
      return Integer.parseInt(s);
  }

  /**
   * get the specified property
   */
  public edu.lehigh.swat.hawk.core.OntProperty getProperty(ObjectId id){
    if(properties.containsKey(id))
      return (edu.lehigh.swat.hawk.core.OntProperty)properties.get(id);
    else if(importedProperties.containsKey(id))
      return (edu.lehigh.swat.hawk.core.OntProperty)importedProperties.get(id);
    return null;

  }

  /**
   * delete a specified property
   */
  public void deleteProperty(ObjectId id){
    properties.remove(id);
  }

  /**
   * get all the properties in this ontology
   */
  public edu.lehigh.swat.hawk.core.OntProperty[] getAllProperties(){
    Collection c = properties.values();
    Object[] oa = c.toArray();
    int len = oa.length;;
    edu.lehigh.swat.hawk.core.OntProperty[] oc = new edu.lehigh.swat.hawk.core.OntProperty[len];
    System.arraycopy(oa,0,oc,0,len);
    return oc;
  }

  /**
   * Get the Iterator of all the properties.
   */
  public Iterator getPropertyIterator(){
    return properties.values().iterator();
  }


  /**
   * add a property to this ontology
   */
  public void addRestriction(edu.lehigh.swat.hawk.core.PropertyRestriction p){
    restrictions.put(p.getID(),p);
  }

  /**
   * get the specified property
   */
  public edu.lehigh.swat.hawk.core.PropertyRestriction getRestriction(ObjectId id){
    if(restrictions.containsKey(id))
      return (edu.lehigh.swat.hawk.core.PropertyRestriction)restrictions.get(id);
    else
      return null;
  }

  /**
   * delete a specified property
   */
  public void deleteRestriction(ObjectId id){
    restrictions.remove(id);
  }

  /**
   * get all the properties in this ontology
   */
  public edu.lehigh.swat.hawk.core.PropertyRestriction[] getAllRestrictions(){
    Collection c = restrictions.values();
    Object[] oa = c.toArray();
    int len = oa.length;
    edu.lehigh.swat.hawk.core.PropertyRestriction[] oc = new edu.lehigh.swat.hawk.core.PropertyRestriction[len];
    System.arraycopy(oa,0,oc,0,len);
    return oc;
  }

  /**
   * Get the priorVersion ontologies which are backwardCompatibleWith
   */
  public DocId[] getCompatiblePriorOnts(){
    if(compatiblePriorOnts.size()<=0)
      return null;
    else{
     Object[] oa = compatiblePriorOnts.toArray();
     int len = oa.length;
     DocId[] oc = new DocId[len];
     System.arraycopy(oa,0,oc,0,len);
     return oc;
    }
  }

  /**
   * Add a priorVersion ontology which is backwardCompatibleWith.
   */
  public void addCompatiblePriorOnts(DocId id){
    compatiblePriorOnts.add(id);
  }

  /**
   * Delete a priorVersion ontology which is backwardCompatibleWith.
   */
  public void deleteCompatiblePriorOnts(DocId id){
    compatiblePriorOnts.remove(id);
  }

  /**
   * Get the priorVersion ontologies may be incompatibleWith
   */
  public DocId[] getPriorOnts(){
    if(priorOnts.size()<=0)
      return null;
    else{
     Object[] oa = priorOnts.toArray();
     int len = oa.length;
     DocId[] oc = new DocId[len];
     System.arraycopy(oa,0,oc,0,len);
     return oc;
    }
  }

  /**
   * Add a priorVersion ontology may be incompatibleWith.
   */
  public void addPriorOnts(DocId id){
    priorOnts.add(id);
  }

  /**
   * Delete a priorVersion ontology may be incompatibleWith.
   */
  public void deletePriorOnts(DocId id){
    priorOnts.remove(id);
  }

  /**
   * get the version of this ontology
   */
  public String getVersion(){
    return version;
  }

  /**
   * Set the version of this ontology.
   */
  public void setVersion(String ver){
    version = ver;
  }

  /**
   * Create a Class object and set its ontology.
   */
  public edu.lehigh.swat.hawk.core.OntClass createClass(){
    OntClass cls = new OntClass();
    cls.setOntology(this);
    return cls;
  }

  /**
   * Create a Property object and set its ontology.
   */
  public edu.lehigh.swat.hawk.core.OntProperty createProperty(){
    OntProperty property = new OntProperty();
    property.setOntology(this);
    return property;

  }

  /**
   * Create a PropertyRestriction object and set its ontology.
   */
  public edu.lehigh.swat.hawk.core.PropertyRestriction createRestriction(){
    PropertyRestriction pr = new PropertyRestriction();
    pr.setOntology(this);
    return pr;
  }

  /**
   * To see if ontology has the class.
   */
  public boolean hasClass(String name){
    Enumeration e = classes.keys();
    while(e.hasMoreElements()){
      ObjectId oid = (ObjectId)e.nextElement();
      if(oid.getLocalName().equalsIgnoreCase(name))
        return true;
    }
    return false;
  }

  /**
   * To see if ontology has the property.
   */
  public boolean hasProperty(String name){
    Enumeration e = properties.keys();
    while(e.hasMoreElements()){
      ObjectId oid = (ObjectId)e.nextElement();
      if(oid.getLocalName().equalsIgnoreCase(name))
        return true;
    }
    return false;
  }

  /**
   * Add an imported class.
   */
  public void addImportedClass(ObjectId cid, edu.lehigh.swat.hawk.core.OntClass c){
    importedClasses.put(cid,c);
  }

  /**
   * Add an imported property.
   */
  public void addImportedProperty(ObjectId pid, edu.lehigh.swat.hawk.core.OntProperty p){
    importedProperties.put(pid,p);
  }


  /**
   * Get All the classes plus the imported Classes.
   */
  public edu.lehigh.swat.hawk.core.OntClass[] getAllPlusImportedClasses(){
    edu.lehigh.swat.hawk.core.OntClass[] ac = this.getAllClasses();
    int num = ac.length;
    Collection c = importedClasses.values();
    Object[] oa = c.toArray();
    int len = oa.length;
    edu.lehigh.swat.hawk.core.OntClass[] oc = new edu.lehigh.swat.hawk.core.OntClass[len+num];
    System.arraycopy(ac,0,oc,0,num);
    System.arraycopy(oa,0,oc,num,len);
    return oc;
  }

  /**
   * Get All the classes plus the imported properties.
   */
  public edu.lehigh.swat.hawk.core.OntProperty[] getAllPlusImportedProperties(){
    edu.lehigh.swat.hawk.core.OntProperty[] ac = this.getAllProperties();
    int num = ac.length;
    Collection c = importedProperties.values();
    Object[] oa = c.toArray();
    int len = oa.length;
    edu.lehigh.swat.hawk.core.OntProperty[] oc = new edu.lehigh.swat.hawk.core.OntProperty[len+num];
    System.arraycopy(ac,0,oc,0,num);
    System.arraycopy(oa,0,oc,num,len);
    return oc;
  }

  /**
   * To tell if the ontology imorted specified class.
   */
  public boolean importedClass(ObjectId cid){
    Enumeration e = importedClasses.keys();
    while(e.hasMoreElements()){
      ObjectId oid = (ObjectId)e.nextElement();
      if(cid.equals(oid))
        return true;
    }
    return false;
  }

  /**
   * To tell if the ontology imported specified property.
   */
  public boolean importedProperty(ObjectId pid){
    Enumeration e = importedProperties.keys();
    while(e.hasMoreElements()){
      ObjectId oid = (ObjectId)e.nextElement();
      if(pid.equals(oid))
        return true;
    }
    return false;
  }

  /**
   * To tell if the ontology asserted specified class.
   */
  public boolean assertedClass(ObjectId cid){
      return assertedClasses.contains(cid);
  }

  /**
   * To tell if the ontology asserted specified property.
   */
  public boolean assertedProperty(ObjectId pid){
      return assertedProperties.contains(pid);
  }

  /**
   * Add an asserted class.
   */
  public void addAssertedClass(ObjectId cid){
      if(!assertedClasses.contains(cid))
          assertedClasses.add(cid);
  }

  /**
   * Add an asserted property.
   */
  public void addAssertedProperty(ObjectId pid){
      if(!assertedProperties.contains(pid))
      assertedProperties.add(pid);
  }


  /**
   * Get the imported class.
   */
//  public edu.lehigh.swat.hawk.core.OntClass getImportedClass(ObjectId id){
//    if(importedClasses.containsKey(id))
//      return (edu.lehigh.swat.hawk.core.OntClass)importedClasses.get(id);
//    else
//      return null;
//  }

  /**
   * Get the imported property.
   */
//  public edu.lehigh.swat.hawk.core.OntProperty getImportedProperty(ObjectId id){
//    if(importedProperties.containsKey(id))
//      return (edu.lehigh.swat.hawk.core.OntProperty)importedProperties.get(id);
//    else
//      return null;
//  }


  private Hashtable classes, properties, restrictions,
  importedClasses, importedProperties;
  private Vector assertedClasses, assertedProperties;
  private Vector priorOnts, compatiblePriorOnts;
  private String version;
  private Vector classIds, propertyIds;
  private Vector<Axiom> axioms;
  private Hashtable<ObjectId, Vector<Axiom>> axiomsBySubject;
  private Hashtable<ObjectId, Vector<Axiom>> axiomsByObject;
  private Hashtable<Integer, Vector<Axiom>> axiomsByType;
  private Vector<AxiomNode> axiomNodeRestrictions;
}
