/*
*Copyright: Copyright (c) 2003-2004, Lehigh University. All rights reserved.
*License Statement: The GNU General Public License
*/


package edu.lehigh.swat.hawk.storage.simplememory;
import edu.lehigh.swat.hawk.storage.StorageModel;
import edu.lehigh.swat.hawk.core.DocId;
import java.util.*;

/**
 * A memory based storage model</p>
 * @author Z. Pan
 * @version 1.0
 */

public class SimpleMemoryStorage extends StorageModel{

  /**
   * Constructor
   */
  public SimpleMemoryStorage() {
    docs = new Hashtable();
  }

  /**
   * Add a document to this model.
   */
  public void addDocument(edu.lehigh.swat.hawk.core.Document doc){
    docs.put(doc.getID().getId(),doc);
  }

  /**
   * Get the specified document.
   */
  public edu.lehigh.swat.hawk.core.Document getDocument(String url){
    if(url.equals(""))
      return null;
    if(docs.containsKey(url))
      return (Document)docs.get(url);
    else
      return null;
  }

  /**
   * Update a document.
   */
  public void updateDocument(edu.lehigh.swat.hawk.core.Document doc){
    if(docs.containsKey(doc.getID().getId()))
      docs.remove(doc.getID().getId());
    docs.put(doc.getID().getId(),doc);
  }

  /**
   *  Get all the documents in the model.
   */
  public Iterator getDocumentIterator(){
    return docs.values().iterator();
  }

  /**
   * Create an Ontology object.
   */
  public edu.lehigh.swat.hawk.core.Ontology createOntology(){
    return new edu.lehigh.swat.hawk.storage.simplememory.Ontology();
  }

  /**
   * Create an Document object.
   */
  public edu.lehigh.swat.hawk.core.Document createDocument(){
    return new edu.lehigh.swat.hawk.storage.simplememory.Document();
  }

  /**
   * Add imports objects to ontology, for the use of memory-based storage model.
   */
  public void pushImports(edu.lehigh.swat.hawk.core.Ontology onto){
    DocId[] iids = onto.getImports();
    if(iids==null)
      return;
    int ilen = iids.length;
    for(int j=0;j<ilen;j++){
      DocId id = iids[j];
      Document idoc = (Document)this.getDocument(id.getId());
      if (! (idoc instanceof Ontology))
        return;
      Ontology ionto = (Ontology) idoc;
      edu.lehigh.swat.hawk.core.OntClass[] ics = ionto.
          getAllPlusImportedClasses();
      int len = ics.length;
      for (int i = 0; i < len; i++) {
        edu.lehigh.swat.hawk.core.OntClass icls = ics[i];
        edu.lehigh.swat.hawk.core.OntClass oc = onto.createClass();
        oc.setID(icls.getID());
        onto.addImportedClass(icls.getID(), oc);
      }
      edu.lehigh.swat.hawk.core.OntProperty[] ips = ionto.
          getAllPlusImportedProperties();
      int plen = ips.length;
      for (int i = 0; i < plen; i++) {
        edu.lehigh.swat.hawk.core.OntProperty ip = ips[i];
        edu.lehigh.swat.hawk.core.OntProperty op = onto.createProperty();
        op.setID(ip.getID());
        onto.addImportedProperty(ip.getID(), op);
      }
    }
  }



  /**
   * Set whether or not include implictly imported documents.
   */
  public void setImplictImports(boolean imp){
    implictImports = imp;
  }

  /**
   *
   * Get the sequence number for giving Document
   */
  public int getSeqNum(DocId did){
    return 0;
  }



  private Hashtable docs;
  private boolean implictImports;
}
